# StarterDeckInfo.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2009 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Adds info about the starter decks cards are found in"""

from sutekh.core.SutekhObjects import PhysicalCardSet, \
        MapPhysicalCardToPhysicalCardSet, IPhysicalCardSet, IRarityPair
from sutekh.core.Filters import PhysicalCardSetFilter, \
        FilterAndBox, SpecificCardIdFilter
from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.CardTextView import CardTextViewListener
from sutekh.gui.ProgressDialog import ProgressDialog, SutekhCountLogHandler
from sutekh.gui.SutekhDialog import SutekhDialog, do_exception_complaint, \
        do_complaint_error
from sutekh.core.CardSetUtilities import delete_physical_card_set, \
        find_children, has_children
from sutekh.io.ZipFileWrapper import ZipFileWrapper
from sutekh.io.DataPack import find_data_pack, DOC_URL
from sutekh.gui.GuiCardSetFunctions import reparent_all_children, \
        update_open_card_sets
from sutekh.gui.FileOrUrlWidget import FileOrUrlWidget, fetch_data
from sutekh.gui.SutekhFileWidget import add_filter
import re
import gtk
import urllib2
from logging import Logger
from StringIO import StringIO
from sqlobject import SQLObjectNotFound


class StarterConfigDialog(SutekhDialog):
    # pylint: disable-msg=R0904
    # R0904 - gtk Widget, so has many public methods
    """Dialog for configuring the Starter plugin."""

    sDocUrl = DOC_URL

    def __init__(self, oParent, bFirstTime=False):
        super(StarterConfigDialog, self).__init__(
                'Configure Starter Info Plugin', oParent,
                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_OK, gtk.RESPONSE_OK, gtk.STOCK_CANCEL,
                    gtk.RESPONSE_CANCEL))
        oDescLabel = gtk.Label()
        if not bFirstTime:
            oDescLabel.set_markup('<b>Choose how to configure the starter info'
                    ' plugin</b>')
        else:
            oDescLabel.set_markup('<b>Choose how to configure the starter info'
                    ' plugin</b>\nChoose cancel to skip configuring the '
                    ' plugin\nYou will not be prompted again')
        self.oFileWidget = FileOrUrlWidget(oParent, "Choose localtion for "
                "Starter decks", {'Sutekh Wiki': self.sDocUrl})
        add_filter(self.oFileWidget, 'Zip Files', ['*.zip', '*.ZIP'])
        # pylint: disable-msg=E1101
        # pylint doesn't pick up vbox methods correctly
        self.vbox.pack_start(oDescLabel, False, False)
        self.vbox.pack_start(self.oFileWidget, False, False)
        # Add check boxes for the decmo and storyline deck questions
        self.oExcludeStoryDecks = gtk.CheckButton(
                'Exclude the Storyline Decks')
        self.oExcludeDemoDecks = gtk.CheckButton(
                'Exclude the WW Demo Decks')

        # Check to see if cards are available
        try:
            _oSkip = IRarityPair(('White Wolf 2003 Demo', 'Demo'))
        except SQLObjectNotFound:
            self.oExcludeDemoDecks.set_active(True)
            self.oExcludeDemoDecks.set_sensitive(False)

        try:
            _oSkip = IRarityPair(('Nergal Storyline', 'Storyline'))
        except SQLObjectNotFound:
            self.oExcludeStoryDecks.set_active(True)
            self.oExcludeStoryDecks.set_sensitive(False)

        self.vbox.pack_start(self.oExcludeStoryDecks, False, False)
        self.vbox.pack_start(self.oExcludeDemoDecks, False, False)

        self.set_size_request(350, 300)

        self.show_all()

    def get_excluded_decks(self):
        """Return selections for excluded decks"""
        return self.oExcludeStoryDecks.get_active(), \
                self.oExcludeDemoDecks.get_active()

    def get_data(self):
        """Return the zip file data containing the decks"""
        sFile, _bUrl = self.oFileWidget.get_file_or_url()
        sData = None
        if sFile == self.sDocUrl:
            # Downloading from sutekh wiki, so need magic to get right file
            sZipUrl = find_data_pack('starters')
            oFile = urllib2.urlopen(sZipUrl)
            sData = fetch_data(oFile)
        elif sFile:
            sData = self.oFileWidget.get_binary_data()
        return sData


def _is_precon(oRarityPair):
    """Return true if the rarity is a precon one"""
    if oRarityPair.rarity.name == 'Precon' or \
            oRarityPair.rarity.name == 'Demo':
        return True
    return False


def _check_precon(oAbsCard):
    """Check if we have at least one precon rarity for this card"""
    for oPair in oAbsCard.rarity:
        if _is_precon(oPair):
            return True
    return False


def _check_exp_name(sExpName, oAbsCard):
    """Check that expansion is one of the precon expansions of the card"""
    for oPair in oAbsCard.rarity:
        if oPair.expansion.name.lower() == sExpName.lower() and \
                _is_precon(oPair):
            return True
    return False


class StarterInfoPlugin(SutekhPlugin, CardTextViewListener):
    """Plugin providing access to CardImageFrame."""
    dTableVersions = {PhysicalCardSet: (5, 6)}
    aModelsSupported = ("MainWindow",)

    dGlobalConfig = {
        'show starters': 'option("Yes", "No", "Unasked", default="Unasked")',
    }

    # FIXME: Expose this to the user?
    oStarterRegex = re.compile('^\[(.*)\] (.*) Starter')
    oDemoRegex = re.compile('^\[(.*)\] (.*) Demo Deck')

    # pylint: disable-msg=W0142
    # ** magic OK here
    def __init__(self, *args, **kwargs):
        super(StarterInfoPlugin, self).__init__(*args, **kwargs)
        self.oToggle = None
        self.oLastCard = None
        self.bShowInfo = False

    def get_menu_item(self):
        """Overrides method from base class.

           Adds the menu item on the MainWindow if the images can be found.
           """
        if not self.check_versions() or not self.check_model_type():
            return None
        # Make sure we add the tag we need
        oCardTextView = self.parent.card_text_pane.view
        oCardTextView.add_listener(self)
        oCardTextView.text_buffer.add_list_tag('starters')

        self.oToggle = gtk.CheckMenuItem("Show Starter Information")
        self.oToggle.connect('toggled', self._toggle_starter)
        self.oToggle.set_active(False)
        if self.check_enabled():
            sPrefsValue = \
                    self.get_config_item('show starters')
            if sPrefsValue == 'Yes':
                self.oToggle.set_active(True)
        else:
            self.oToggle.set_sensitive(False)
        oDownload = gtk.MenuItem("Download starter decks")
        oDownload.connect('activate', self.do_download)
        return [('Preferences', self.oToggle), ('Data Downloads', oDownload)]

    def check_enabled(self):
        """check for starter decks in the database and disable menu if not"""
        bEnabled = False
        for oCS in PhysicalCardSet.select():
            oMatch = self.oStarterRegex.match(oCS.name)
            if oMatch:
                bEnabled = True
                break
        return bEnabled

    def setup(self):
        """1st time setup tasks"""
        sPrefsValue = self.get_config_item('show starters')
        if sPrefsValue == 'Unasked':
            self.set_config_item('show starters', 'No')
            if not self.check_enabled():
                # First time
                oDialog = StarterConfigDialog(self.parent, True)
                self.handle_response(oDialog)

    def do_download(self, _oMenuWidget):
        """Prompt the user to download/setup decks"""
        oDialog = StarterConfigDialog(self.parent)
        self.handle_response(oDialog)

    def handle_response(self, oDialog):
        """Handle the response from the config dialog"""
        iResponse = oDialog.run()
        if iResponse == gtk.RESPONSE_OK:
            sData = oDialog.get_data()
            (bExcludeStoryDecks, bExcludeDemoDecks) = \
                    oDialog.get_excluded_decks()
            if not sData:
                do_complaint_error('Unable to access zipfile data')
            elif not self._unzip_file(sData, bExcludeStoryDecks,
                    bExcludeDemoDecks):
                do_complaint_error('Unable to successfully unzip zipfile')
        if self.check_enabled():
            self.oToggle.set_sensitive(True)
        else:
            self.oToggle.set_sensitive(False)
        # cleanup
        oDialog.destroy()

    def _unzip_file(self, sData, bExcludeStoryDecks, bExcludeDemoDecks):
        """Unzip a file containing the decks."""
        bResult = False
        oZipFile = ZipFileWrapper(StringIO(sData))
        bResult = self._unzip_heart(oZipFile, bExcludeStoryDecks,
                bExcludeDemoDecks)
        return bResult

    def _unzip_heart(self, oFile, bExcludeStoryDecks, bExcludeDemoDecks):
        """Heart of the reading loop - ensure we read parents before
           children, and correct for renames that occur."""
        oLogHandler = SutekhCountLogHandler()
        oProgressDialog = ProgressDialog()
        oProgressDialog.set_description("Importing Starters")
        oLogger = Logger('Read zip file')
        aExistingList = [x.name for x in PhysicalCardSet.select()]
        dList = oFile.get_all_entries()
        # Check that we match starter regex
        bOK = False
        for sName in dList:
            oMatch = self.oStarterRegex.match(sName)
            if oMatch:
                bOK = True
                break
        if not bOK:
            oProgressDialog.destroy()
            return False  # No starters in zip file
        oLogger.addHandler(oLogHandler)
        oLogHandler.set_dialog(oProgressDialog)
        oLogHandler.set_total(len(dList))
        oProgressDialog.show()
        bDone = False
        while not bDone:
            dRemaining = {}
            if self._unzip_list(oFile, dList, oLogger, dRemaining,
                    bExcludeStoryDecks, bExcludeDemoDecks):
                bDone = len(dRemaining) == 0
                dList = dRemaining
            else:
                self.reload_pcs_list()
                oProgressDialog.destroy()
                return False  # Error
        # Cleanup
        self._clean_empty(oFile.get_all_entries(), aExistingList)
        self.reload_pcs_list()
        oProgressDialog.destroy()
        return True

    def _unzip_list(self, oZipFile, dList, oLogger, dRemaining,
            bExcludeStoryDecks, bExcludeDemoDecks):
        """Extract the files left in the list."""
        if bExcludeStoryDecks and bExcludeDemoDecks:
            aExcluded = ["White Wolf Storyline Decks", "White Wolf Demo Decks"]
        elif bExcludeStoryDecks:
            aExcluded = ["White Wolf Storyline Decks"]
        elif bExcludeDemoDecks:
            aExcluded = ["White Wolf Demo Decks"]
        else:
            aExcluded = []
        for sName, tInfo in dList.iteritems():
            sFilename, _bIgnore, sParentName = tInfo
            if sName in aExcluded or sParentName in aExcluded:
                # Ignore these decks
                oLogger.info('Read %s' % sName)
                continue
            if sParentName is not None and sParentName in dList:
                # Do have a parent to look at later, so skip for now
                dRemaining[sName] = tInfo
                continue
            elif sParentName is not None:
                if PhysicalCardSet.selectBy(name=sParentName).count() == 0:
                    # Missing parent, so it the file is invalid
                    return False
            # pylint: disable-msg=W0703
            # we really do want all the exceptions
            try:
                oHolder = oZipFile.read_single_card_set(sFilename)
                oLogger.info('Read %s' % sName)
                if not oHolder.name:
                    # We skip this card set
                    continue
                # We unconditionally delete the card set if it already
                # exists, as being the most sensible default
                aChildren = []
                if PhysicalCardSet.selectBy(name=oHolder.name).count() != 0:
                    # pylint: disable-msg=E1101, E1103
                    # pyprotocols confuses pylint
                    oCS = IPhysicalCardSet(oHolder.name)
                    aChildren = find_children(oCS)
                    # Ensure we restore with the correct parent
                    if oCS.parent:
                        oHolder.parent = oCS.parent.name
                    else:
                        oHolder.parent = None
                    delete_physical_card_set(oHolder.name)
                oHolder.create_pcs(self.cardlookup)
                reparent_all_children(oHolder.name, aChildren)
                if self.parent.find_cs_pane_by_set_name(oHolder.name):
                    # Already open, so update to changes
                    update_open_card_sets(self.parent, oHolder.name)
            except Exception, oException:
                sMsg = "Failed to import card set %s.\n\n%s" % (sName,
                        oException)
                do_exception_complaint(sMsg)
                return False
        return True

    # pylint: disable-msg=R0201
    # Method for consistency with _unzip methods
    def _clean_empty(self, aMyList, aExistingList):
        """Remove any newly created sets in that have no cards AND no
           children"""
        for sName in aMyList:
            if sName in aExistingList:
                continue  # Not a card set we added
            try:
                oCS = IPhysicalCardSet(sName)
            except SQLObjectNotFound:
                # set not there, so skip
                continue
            # pylint: disable-msg=E1101
            # QLObject + PyProtocols confuses pylint
            if has_children(oCS):
                continue
            if len(oCS.cards) > 0:
                continue  # has cards
            delete_physical_card_set(sName)

    # pylint: enable-msg=R0201

    def _toggle_starter(self, oToggle):
        """Toggle the show info flag"""
        self.bShowInfo = oToggle.get_active()
        # Update the card text pane to reflect changes
        self.parent.set_card_text(self.oLastCard)
        if self.bShowInfo:
            self.set_config_item('show starters', 'Yes')
        else:
            self.set_config_item('show starters', 'No')

    def _get_card_set_info(self, oAbsCard):
        """Find preconstructed card sets that contain the card"""
        dMatches = {'Starters': [], 'Demos': []}
        for oCS in PhysicalCardSet.select():
            for sType, oRegex in (('Starters', self.oStarterRegex),
                    ('Demos', self.oDemoRegex)):
                oMatch = oRegex.match(oCS.name)
                if oMatch:
                    sExpName = oMatch.groups()[0]
                    if _check_exp_name(sExpName, oAbsCard):
                        dMatches[sType].append((oCS, oMatch.groups()[0],
                            oMatch.groups()[1]))
        # Find the card in the starter decks
        dInfo = {'Starters': [], 'Demos': []}
        for sType, aResults in dMatches.iteritems():
            for oCS, sExpName, sDeckName in sorted(aResults,
                    key=lambda x: (x[1], x[2])):
                # Sort by exp, name
                oFilter = FilterAndBox([SpecificCardIdFilter(oAbsCard.id),
                        PhysicalCardSetFilter(oCS.name)])
                iCount = oFilter.select(
                        MapPhysicalCardToPhysicalCardSet).count()
                if iCount > 0:
                    dInfo[sType].append("x %(count)d %(exp)s (%(cardset)s)" % {
                        'count': iCount,
                        'exp': sExpName,
                        'cardset': sDeckName,
                        })
        return dInfo

    def set_card_text(self, oPhysCard):
        """Update the card text pane with the starter info"""
        self.oLastCard = oPhysCard
        if not self.bShowInfo:
            return  # Do nothing
        oAbsCard = oPhysCard.abstractCard
        if not _check_precon(oAbsCard):
            return
        # Find the starter decks
        dInfo = self._get_card_set_info(oAbsCard)
        # Get frame
        oCardTextBuf = self.parent.card_text_pane.view.text_buffer
        oTempIter = oCardTextBuf.get_cur_iter()
        # We do things in this order, so demo decks appear after the starters
        if dInfo['Demos']:
            # Move to after the expansionsa
            oCardTextBuf.set_cur_iter_to_mark('expansion')
            oCardTextBuf.tag_text("\n")
            oCardTextBuf.labelled_list('Demonstration Decks', dInfo['Demos'],
                    'starters')
        if dInfo['Starters']:
            oCardTextBuf.set_cur_iter_to_mark('expansion')
            oCardTextBuf.tag_text("\n")
            oCardTextBuf.labelled_list('Preconstructed Decks',
                    dInfo['Starters'], 'starters')
        oCardTextBuf.set_cur_iter(oTempIter)

plugin = StarterInfoPlugin



# CachedRelatedJoin.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2007 Simon Cross <hodgestar@gmail.com>,
# Copyright 2007 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Implement RelatedJoin with caches"""

from sqlobject import joins
from sqlobject.sqlbuilder import Table, Select


class SOCachedRelatedJoin(joins.SORelatedJoin):
    """Version of RelatedJoin that caches the lookup of related objects.

       Updates to the database require explicitly flushing the cache on
       each join. E.g.:

       for oJoin in AbstractCard.sqlmeta.joins:
         if type(oJoin) is SOCachedRelatedJoin:
            oJoin.flush_cache()
       """

    # pylint: disable-msg=W0142
    # ** magic OK here
    def __init__(self, *aArgs, **kwargs):
        super(SOCachedRelatedJoin, self).__init__(*aArgs, **kwargs)
        self._dJoinCache = {}
        self._oOtherJoin = None
        self._bOtherJoinCached = None

    def _find_other_join(self):
        """Locate the equivalent join on the other class."""
        if self._oOtherJoin is None:
            aJoins = [oJ for oJ in self.otherClass.sqlmeta.joins
                        if oJ.otherClass is self.soClass]
            assert len(aJoins) == 1
            self._oOtherJoin = aJoins[0]

            self._bOtherJoinCached = isinstance(self._oOtherJoin,
                                                SOCachedRelatedJoin)

    def flush_cache(self):
        """Flush the contents of the cache."""
        self._dJoinCache = {}

    def init_cache(self):
        """Initialise the cache with the data from the database."""
        self._find_other_join()

        oIntermediateTable = Table(self.intermediateTable)
        oJoinColumn = getattr(oIntermediateTable, self.joinColumn)
        oOtherColumn = getattr(oIntermediateTable, self.otherColumn)
        # pylint: disable-msg=W0212
        # We need to access _connection here
        oConn = self.soClass._connection

        for (oId, oOtherId) in oConn.queryAll(repr(Select(
            (oJoinColumn, oOtherColumn)))):
            oInst = self.soClass.get(oId, oConn)
            oOther = self.otherClass.get(oOtherId, oConn)
            self._dJoinCache.setdefault(oInst, [])
            self._dJoinCache[oInst].append(oOther)

        # Apply ordering (we assume it won't change later)
        for oInst in self._dJoinCache:
            self._dJoinCache[oInst] = self._applyOrderBy(
                    self._dJoinCache[oInst], self.otherClass)

    def invalidate_cache_item(self, oInst, oOther, bDoOther=True):
        """Invalidate a cache item and its equivalent in the other join."""
        if oInst in self._dJoinCache:
            del self._dJoinCache[oInst]
        if bDoOther and self._bOtherJoinCached:
            self._find_other_join()
            self._oOtherJoin.invalidate_cache_item(oOther, oInst,
                                                   bDoOther=False)

    # pylint: disable-msg=C0103
    # Name must match SQLObject conventions
    def performJoin(self, oInst):
        """Return the join the result, from the cache if possible."""
        if not oInst in self._dJoinCache:
            self._dJoinCache[oInst] = joins.SORelatedJoin.performJoin(self,
                    oInst)
        return self._dJoinCache[oInst]

    def add(self, oInst, oOther):
        """Add an item to the join."""
        self.invalidate_cache_item(oInst, oOther)
        super(SOCachedRelatedJoin, self).add(oInst, oOther)

    def remove(self, oInst, oOther):
        """Remove an item from the join."""
        self.invalidate_cache_item(oInst, oOther)
        super(SOCachedRelatedJoin, self).remove(oInst, oOther)


class CachedRelatedJoin(joins.RelatedJoin):
    """Provide CacheRelatedJoin object to Sutekh"""
    baseClass = SOCachedRelatedJoin



# WriteELDBInventory.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Writer for the FELDB inventory format.

   Example:

   "ELDB - Inventory"
   "Aabbt Kindred",1,0,"","Crypt"
   "Aaron Duggan, Cameron`s Toady",0,0,"","Crypt"
   ...
   "Zip Line",0,0,"","Library"
   """

from sutekh.core.SutekhObjects import AbstractCard, IAbstractCard
from sutekh.core.ELDBUtilities import norm_name, type_of_card


class WriteELDBInventory(object):
    """Create a string in ELDB inventory format representing a card set."""

    # pylint: disable-msg=R0201
    # method for consistency with the other methods
    def _gen_header(self):
        """Generate an ELDB inventory file header."""
        return '"ELDB - Inventory"'

    def _gen_inv(self, oHolder):
        """Process the card set, creating the lines as needed"""
        dCards = {}
        sResult = ""
        for oCard in AbstractCard.select():
            dCards[oCard] = 0
        for oCard in oHolder.cards:
            oAbsCard = IAbstractCard(oCard)
            dCards[oAbsCard] += 1
        for oCard, iNum in dCards.iteritems():
            sResult += '"%s",%d,0,"","%s"\n' % (norm_name(oCard), iNum,
                    type_of_card(oCard))
        return sResult

    # pylint: enable-msg=R0201

    def write(self, fOut, oHolder):
        """Takes file object + card set to write, and writes an ELDB inventory
           representing the deck"""
        fOut.write(self._gen_header())
        fOut.write("\n")
        fOut.write(self._gen_inv(oHolder))



# SutekhIcon.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Provide a class for handling the application icon"""

# pylint: disable-msg=E0611
# pylint doesn't see resource_stream here, for some reason
from pkg_resources import resource_stream
# pylint: enable-msg=E0611
import gtk


class PkgIcon(object):
    """Load a gtk Pixbuf object from a package resource file."""

    def __init__(self, sPkg, sResource):
        oLoader = gtk.gdk.PixbufLoader()
        oFile = resource_stream(sPkg, sResource)
        oLoader.write(oFile.read())
        oFile.close()
        oLoader.close()
        self._oIcon = oLoader.get_pixbuf()

    def get_pixbuf(self):
        """Return the actual icon"""
        return self._oIcon

# pylint: disable-msg=C0103
# special case, so convention doesn't apply
SUTEKH_ICON = PkgIcon(__name__, "sutekh-icon.svg").get_pixbuf()



# CountWWListCards.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Simon Cross <hodgestar@gmail.com>,
# Copyright 2006 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Provide count card options for the WW card list"""

import gtk
from sutekh.core.SutekhObjects import PhysicalCard, IAbstractCard
from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.CardListModel import CardListModelListener
from sutekh.SutekhUtility import is_crypt_card
from sutekh.gui.plugins.CountCardSetCards import TOT_FORMAT, TOT_TOOLTIP, \
        TOTAL, CRYPT, LIB

SORT_COLUMN_OFFSET = 300  # ensure we don't clash with other extra columns


class CountWWListCards(SutekhPlugin, CardListModelListener):
    """Listen to changes on the card list views, and display a toolbar
       containing a label with a running count of the cards in the card
       set, the library cards and the crypt cards
       """
    dTableVersions = {PhysicalCard: (2,)}
    aModelsSupported = (PhysicalCard,)

    NO_COUNT, COUNT_CARDS, COUNT_EXP = range(3)

    COLUMN_NAME = '#'

    NO_COUNT_OPT = "Don't show card counts"
    COUNT_CARD_OPT = "Show counts for each distinct card"
    COUNT_EXP_OPT = "Show counts for each expansion"

    MODES = {
            NO_COUNT_OPT: NO_COUNT,
            COUNT_CARD_OPT: COUNT_CARDS,
            COUNT_EXP_OPT: COUNT_EXP,
            }

    OPTION_NAME = "White Wolf Card List Count Mode"

    OPTION_STR = ", ".join('"%s"' % sKey for sKey in sorted(MODES.keys()))

    dCardListConfig = {
        OPTION_NAME: 'option(%s, default="%s")' % (OPTION_STR, NO_COUNT_OPT),
    }

    # pylint: disable-msg=W0142
    # **magic OK here
    def __init__(self, *args, **kwargs):
        super(CountWWListCards, self).__init__(*args, **kwargs)

        self._oTextLabel = None
        self._iMode = self.NO_COUNT_OPT
        self._dExpCounts = {}
        self._dAbsCounts = {}
        self._dCardTotals = {TOTAL: 0, CRYPT: 0, LIB: 0}
        self._dExpTotals = {TOTAL: 0, CRYPT: 0, LIB: 0}

        # We only add listeners to windows we're going to display the toolbar
        # on
        if self.check_versions() and self.check_model_type():
            self.model.add_listener(self)
    # pylint: enable-msg=W0142

    def cleanup(self):
        """Remove the listener"""
        if self.check_versions() and self.check_model_type():
            self.model.remove_listener(self)
        super(CountWWListCards, self).cleanup()

    def _get_card_count(self, oAbsCard):
        """Get the count for the card for the current mode"""
        if self._iMode == self.COUNT_EXP:
            # We default if we're only showing the unspecified expansion
            return self._dExpCounts.get(oAbsCard, 0)
        else:
            return self._dAbsCounts[oAbsCard]

    def get_toolbar_widget(self):
        """Overrides method from base class."""
        if not self.check_versions() or not self.check_model_type():
            return None

        dInfo = {TOTAL: 0, CRYPT: 0, LIB: 0}
        self._oTextLabel = gtk.Label(TOT_FORMAT % dInfo)
        self._oTextLabel.set_tooltip_markup(TOT_TOOLTIP % dInfo)

        if self._iMode != self.NO_COUNT:
            self._oTextLabel.show()
        else:
            self._oTextLabel.hide()
        return self._oTextLabel

    def update_numbers(self):
        """Update the label"""
        # Timing issues mean that this can be called before text label has
        # been properly realised, so we need this guard case
        if self._oTextLabel:
            if self._iMode == self.NO_COUNT:
                self._oTextLabel.hide()
                return
            elif self._iMode == self.COUNT_CARDS:
                dInfo = self._dCardTotals
            else:
                dInfo = self._dExpTotals
            self._oTextLabel.set_markup(TOT_FORMAT % dInfo)
            self._oTextLabel.set_tooltip_markup(TOT_TOOLTIP % dInfo)
            self._oTextLabel.show()

    def load(self, aCards):
        """Listen on load events & update counts"""
        # The logic is a bit complicated, but it's intended that
        # filtering the WW cardlist on a card set will give sensible
        # results.
        self._dAbsCounts = {}
        self._dExpCounts = {}
        self._dCardTotals = {TOTAL: 0, CRYPT: 0, LIB: 0}
        self._dExpTotals = {TOTAL: 0, CRYPT: 0, LIB: 0}
        for oCard in aCards:
            oAbsCard = IAbstractCard(oCard)
            if oAbsCard not in self._dAbsCounts:
                self._dAbsCounts[oAbsCard] = 1
                iAbsCount = 1
            else:
                iAbsCount = 0
            if oCard.expansionID:
                # We don't count expansion ifno for cards with no expansion set
                iExpCount = 1
                if oAbsCard not in self._dExpCounts:
                    # First time we've seen this card
                    self._dExpCounts[oAbsCard] = 1
                else:
                    # Has an expansion, and by the nature of the list, this is
                    # a new expansion for the card we haven't seen before
                    self._dExpCounts[oAbsCard] += 1
            else:
                iExpCount = 0
            if is_crypt_card(oAbsCard):
                self._dExpTotals[CRYPT] += iExpCount
                self._dCardTotals[CRYPT] += iAbsCount
            else:
                self._dExpTotals[LIB] += iExpCount
                self._dCardTotals[LIB] += iAbsCount
            self._dCardTotals[TOTAL] += iAbsCount
            self._dExpTotals[TOTAL] += iExpCount
        self.update_numbers()

    def perpane_config_updated(self, _bDoReload=True):
        """Called by base class on config updates."""
        if self.check_versions() and self.check_model_type():
            sCountMode = self.get_perpane_item(self.OPTION_NAME)
            self._iMode = self.MODES.get(sCountMode, self.NO_COUNT)
            if self._iMode == self.NO_COUNT:
                self.clear_col()
            else:
                self.add_col()
            self.update_numbers()

    def _get_cols(self):
        """Get a list holding the column"""
        return [oCol for oCol in self.view.get_columns() if
                oCol.get_property("title") == self.COLUMN_NAME]

    def clear_col(self):
        """Remove the column if it's in use"""
        for oCol in self._get_cols():
            self.view.remove_column(oCol)

    def add_col(self):
        """Add the count col"""
        aCols = self._get_cols()
        if aCols:
            return  # column already there
        oCell = gtk.CellRendererText()
        oColumn = gtk.TreeViewColumn(self.COLUMN_NAME, oCell)
        oColumn.set_cell_data_func(oCell, self._render_count)
        oColumn.set_sizing(gtk.TREE_VIEW_COLUMN_FIXED)
        oColumn.set_fixed_width(40)
        oColumn.set_resizable(True)
        self.view.insert_column(oColumn, 0)
        oColumn.set_sort_column_id(SORT_COLUMN_OFFSET)
        self.model.set_sort_func(SORT_COLUMN_OFFSET, self._sort_count)

    def _get_count(self, oIter):
        """Get the count for this iter"""
        if self.model.iter_depth(oIter) == 0:
            # Group level item, so need to total counts for all subiters
            iTot = 0
            oSubIter = self.model.iter_children(oIter)
            while oSubIter:
                iTot += self._get_count(oSubIter)
                oSubIter = self.model.iter_next(oSubIter)
            return iTot
        elif self.model.iter_depth(oIter) == 1:
            oAbsCard = self.model.get_abstract_card_from_iter(oIter)
            return self._get_card_count(oAbsCard)
        elif self.model.iter_depth(oIter) == 2 and \
                self._iMode == self.COUNT_EXP:
            oPhysCard = self.model.get_physical_card_from_iter(oIter)
            if oPhysCard.expansionID:
                return 1
        return 0

    def _render_count(self, _oColumn, oCell, _oModel, oIter):
        """Render the count for the card"""
        iVal = self._get_count(oIter)
        oCell.set_property('text', iVal)

    def _sort_count(self, _oModel, oIter1, oIter2):
        """Card count Comparision of oIter1 and oIter2.

           Return -1 if oIter1 < oIter, 0 in ==, 1 if >
           """
        iVal1 = self._get_count(oIter1)
        iVal2 = self._get_count(oIter2)
        if iVal1 < iVal2:
            return -1
        elif iVal1 > iVal2:
            return 1
        else:
            # Values agree, so do fall back sort
            return self.model.sort_equal_iters(oIter1, oIter2)


plugin = CountWWListCards



# CardSetIndependence.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Simon Cross <hodgestar@gmail.com>,
# Copyright 2006 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Test whether card sets can be constructed independently"""

import gtk
from sutekh.core.SutekhObjects import PhysicalCardSet, IPhysicalCardSet, \
        IAbstractCard
from sutekh.core.Filters import ParentCardSetFilter
from sutekh.gui.PluginManager import SutekhPlugin
from sutekh.gui.CardSetsListView import CardSetsListView
from sutekh.gui.SutekhDialog import SutekhDialog, do_complaint, \
        do_complaint_error
from sutekh.gui.AutoScrolledWindow import AutoScrolledWindow


# helper functions and classes
class CardInfo(object):
    """Helper class to hold card set info"""
    def __init__(self):
        self.iCount = 0
        self.dCardSets = {}

    def format_cs(self):
        """Pretty print card set list"""
        return ", ".join(self.dCardSets)


def _get_cards(oCardSet, dCards, bIgnoreExpansions):
    """Extract the abstract cards from the card set oCardSet"""
    # pylint: disable-msg=E1101
    # SQLObject + pyprotocol methods confuse pylint
    if bIgnoreExpansions:
        fCard = lambda oCard: oCard.abstractCard
    else:
        fCard = lambda oCard: oCard
    for oCard in oCardSet.cards:
        oCard = fCard(oCard)
        dCards.setdefault(oCard, CardInfo())
        dCards[oCard].iCount += 1
        dCards[oCard].dCardSets.setdefault(oCardSet.name, 0)
        dCards[oCard].dCardSets[oCardSet.name] += 1


def _make_align_list(aList):
    """Wrap the list of strings in an aligned widget for display."""
    oLabel = gtk.Label()
    oAlign = gtk.Alignment()
    oAlign.add(oLabel)
    oAlign.set_padding(0, 0, 5, 0)  # offset a little from the left edge
    sContents = "\n".join(aList)
    oLabel.set_markup(sContents)
    oLabel.set_selectable(True)
    return oAlign


class CardSetIndependence(SutekhPlugin):
    """Provides a plugin for testing whether card sets are independant.

       Independence in this cases means that there are enought cards in
       the parent card set to construct all the selected child card sets
       simulatenously.

       We don't test the case when parent is None, since there's nothing
       particularly sensible to say there. We also don't do anything
       when there is only 1 child, for similiar justification.
       """
    dTableVersions = {PhysicalCardSet: (4, 5, 6)}
    aModelsSupported = (PhysicalCardSet,)

    def get_menu_item(self):
        """Register with the 'Analyze' Menu"""
        if not self.check_versions() or not self.check_model_type():
            return None
        oTest = gtk.MenuItem("Test Card Set Independence")
        oTest.connect("activate", self.activate)
        return ('Analyze', oTest)

    def activate(self, _oWidget):
        """Create the dialog in response to the menu item."""
        oDlg = self.make_dialog()
        if oDlg:
            oDlg.run()

    def make_dialog(self):
        """Create the list of card sets to select"""
        # pylint: disable-msg=W0201, E1101
        # E1101: PyProtocols confuses pylint
        # W0201: No need to define oThisCardSet, oCSView & oInUseButton in
        # __init__
        self.oThisCardSet = IPhysicalCardSet(self.view.sSetName)
        if not self.oThisCardSet.parent:
            do_complaint_error("Card Set has no parent, so nothing to test.")
            return None

        bInUseSets = PhysicalCardSet.selectBy(
                parentID=self.oThisCardSet.parent.id, inuse=True).count() > 0
        oDlg = SutekhDialog("Choose Card Sets to Test", self.parent,
                          gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                          (gtk.STOCK_OK, gtk.RESPONSE_OK,
                           gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL))
        self.oCSView = CardSetsListView(None, oDlg)
        # pylint: disable-msg=E1101
        # vbox confuses pylint
        oDlg.vbox.pack_start(AutoScrolledWindow(self.oCSView), expand=True)
        self.oCSView.set_select_multiple()
        self.oCSView.exclude_set(self.view.sSetName)
        # exclude all parents
        oParentSet = self.oThisCardSet.parent
        while oParentSet:
            self.oCSView.exclude_set(oParentSet.name)
            oParentSet = oParentSet.parent
        # Add filter so we only show the sibling card sets
        oFilter = ParentCardSetFilter([self.oThisCardSet.parent.name])
        self.oCSView.set_filter(oFilter, None)
        self.oCSView.load()
        self.oCSView.expand_to_entry(self.view.sSetName)
        self.oCSView.set_size_request(450, 300)
        self.oIgnoreExpansions = gtk.CheckButton(
                label="Ignore card expansions")
        oDlg.vbox.pack_start(self.oIgnoreExpansions, False, False)
        self.oInUseButton = gtk.CheckButton(label="Test against all cards sets"
                " marked as in use")
        if bInUseSets:
            oDlg.vbox.pack_start(self.oInUseButton, False, False)
            self.oInUseButton.connect("toggled", self.show_hide_list)
        oDlg.connect("response", self.handle_response)
        oDlg.show_all()
        return oDlg

    def show_hide_list(self, oWidget):
        """Hide the card set selection when the user toggles the check box"""
        if oWidget.get_active():
            self.oCSView.set_select_none()
        else:
            self.oCSView.set_select_multiple()

    def handle_response(self, oDlg, oResponse):
        """Handle the response from the dialog."""
        # pylint: disable-msg=E1101
        # Pyprotocols confuses pylint
        if oResponse == gtk.RESPONSE_OK:
            bIgnoreExpansions = self.oIgnoreExpansions.get_active()
            if self.oInUseButton.get_active():
                oInUseSets = PhysicalCardSet.selectBy(
                        parentID=self.oThisCardSet.parent.id, inuse=True)
                aCardSetNames = [oCS.name for oCS in oInUseSets]
                if self.view.sSetName not in aCardSetNames:
                    aCardSetNames.append(self.view.sSetName)
            else:
                aCardSetNames = [self.view.sSetName]
                aSelected = self.oCSView.get_all_selected_sets()
                if aSelected:
                    aCardSetNames.extend(aSelected)
            # We may have only 1 set here, but we will still do the
            # right thing in that case.
            self._test_card_sets(aCardSetNames, self.oThisCardSet.parent,
                    bIgnoreExpansions)
        oDlg.destroy()

    def _display_results(self, dMissing, oParentCS):
        """Display the list of missing cards"""
        # pylint: disable-msg=E1101
        # E1101: PyProtocols confuses pylint
        oResultDlg = SutekhDialog("Missing Cards", None,
                gtk.DIALOG_MODAL | gtk.DIALOG_DESTROY_WITH_PARENT,
                (gtk.STOCK_CLOSE, gtk.RESPONSE_CLOSE))
        oNotebook = gtk.Notebook()
        oNotebook.set_scrollable(True)
        oNotebook.popup_enable()
        dFormatted = {}
        aParentList = []
        for oCard, oInfo in dMissing.iteritems():
            sCardName = self.escape(IAbstractCard(oCard).name)
            if not hasattr(oCard, 'expansion'):
                sExpansion = ""
            elif oCard.expansion:
                sExpansion = " [%s]" % self.escape(oCard.expansion.name)
            else:
                sExpansion = " [(unspecified expansion)]"
            aParentList.append('<span foreground = "blue">'
                    '%s%s</span> : Missing <span foreground="red">%d'
                    '</span> copies (used in %s)'
                    % (sCardName, sExpansion, oInfo.iCount,
                        self.escape(oInfo.format_cs())))
            for sCardSet, iCount in oInfo.dCardSets.iteritems():
                dFormatted.setdefault(sCardSet, [])
                dFormatted[sCardSet].append('<span foreground="blue">%s%s'
                        '</span>: Missing <span foreground="red">%d'
                        '</span> copies (%d copies used here)' % (sCardName,
                            sExpansion, oInfo.iCount, iCount))
        oHeading = gtk.Label()
        oHeading.set_markup('Missing from %s' % self.escape(oParentCS.name))
        oNotebook.append_page(AutoScrolledWindow(
            _make_align_list(aParentList), True), oHeading)
        for sCardSet, aMsgs in dFormatted.iteritems():
            oHeading = gtk.Label()
            oHeading.set_markup('Missing in %s' % self.escape(sCardSet))
            oNotebook.append_page(AutoScrolledWindow(
                _make_align_list(aMsgs), True), oHeading)
        # pylint: disable-msg=E1101
        # pylint misses vbox methods
        oResultDlg.vbox.pack_start(oNotebook)
        oResultDlg.set_size_request(600, 600)
        oResultDlg.show_all()
        oResultDlg.run()
        oResultDlg.destroy()

    def _test_card_sets(self, aCardSetNames, oParentCS, bIgnoreExpansions):
        """Test if the Card Sets are actaully independent by
           looking for cards common to the sets"""
        dCards = {}
        for sCardSetName in aCardSetNames:
            oCS = IPhysicalCardSet(sCardSetName)
            _get_cards(oCS, dCards, bIgnoreExpansions)
        dParent = {}
        _get_cards(oParentCS, dParent, bIgnoreExpansions)
        dMissing = {}
        for oCard, oInfo in dCards.iteritems():
            if oCard not in dParent:
                dMissing[oCard] = oInfo
            elif dParent[oCard].iCount < oInfo.iCount:
                # the dict ensures we don't need a copy
                dMissing[oCard] = oInfo
                dMissing[oCard].iCount -= dParent[oCard].iCount
        if dMissing:
            self._display_results(dMissing, oParentCS)
        else:
            sMessage = "No cards missing from %s" % oParentCS.name
            do_complaint(sMessage, gtk.MESSAGE_INFO, gtk.BUTTONS_CLOSE, True)


plugin = CardSetIndependence



# FilteredViewMenu.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Base class for the frame menu's
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Base class for the pane menus"""

from sutekh.gui.SutekhMenu import SutekhMenu
import gtk


class FilteredViewMenu(SutekhMenu):
    # pylint: disable-msg=R0904, R0922
    # R0904 - gtk.Widget, so many public methods
    # R0922 - we use this in other files
    """Base class for individual FilteredView menus

       This provides handling for enabling and disabling the menus
       ability to receive accelerators on focus changes, expanding
       & collapsing rows, and manging the active filter.
       """
    def __init__(self, oFrame, oWindow, oController):
        super(FilteredViewMenu, self).__init__(oWindow)
        self._oFrame = oFrame
        self.oApply = None
        self._oController = oController

    def cleanup(self):
        """Cleanup hook for menus"""
        pass

    def create_filter_menu(self):
        """Create the Filter Menu."""
        oMenu = self.create_submenu(self, "F_ilter")
        self.create_menu_item("_Specify Filter", oMenu, self.set_active_filter,
                '<Ctrl>s')
        self.oApply = self.create_check_menu_item("_Apply Filter", oMenu,
                self.toggle_apply_filter, False, "<Ctrl>t")
        self.oApply.set_inconsistent(False)

    def create_edit_menu(self):
        """Create the 'Edit' menu, and populate it."""
        oMenu = self.create_submenu(self, "_Edit")
        self.add_edit_menu_actions(oMenu)

    def set_active_filter(self, _oWidget):
        """Set the current filter for the card set."""
        self._oController.view.get_filter(self)

    def toggle_apply_filter(self, oWidget):
        """Toggle the filter applied state."""
        self._oController.view.run_filter(oWidget.active)

    def set_apply_filter(self, bState):
        """Set the applied filter state to bState."""
        self.oApply.set_active(bState)

    def get_apply_filter(self):
        """Get the filter applied state"""
        return self.oApply.active

    def add_common_actions(self, oMenu):
        """Actions common to all card lists"""
        self.create_menu_item("Expand All", oMenu, self.expand_all,
                '<Ctrl>plus')
        self.create_menu_item("Collapse All", oMenu, self.collapse_all,
                '<Ctrl>minus')

    def add_edit_menu_actions(self, oMenu):
        """Add the search item to the Edit Menu."""
        self.create_menu_item('_Search', oMenu, self.show_search_dialog,
                '<Ctrl>f')

    def expand_all(self, _oWidget):
        """Expand all the rows in the card set."""
        self._oController.view.expand_all()

    def collapse_all(self, _oWidget):
        """Collapse all the rows in the card set."""
        self._oController.view.collapse_all()

    def show_search_dialog(self, _oWidget):
        """Show the search dialog"""
        self._oController.view.emit('start-interactive-search')

    # profile helpers

    # pylint: disable-msg=R0913
    # We need all these arguments
    def _create_profile_menu(self, oParentMenu, sTitle, sType, fCallback,
            sProfile):
        """Create a radio group sub-menu for selecting a profile."""
        oMenu = self.create_submenu(oParentMenu, sTitle)
        oConfig = self._oMainWindow.config_file

        oGroup = gtk.RadioMenuItem(None,
            oConfig.get_profile_option(sType, None, "name"))
        oGroup.connect("toggled", fCallback, None)
        oMenu.append(oGroup)

        self._update_profile_group(oMenu, sType, fCallback, sProfile)

        return oMenu

    # pylint: enable-msg=R0913

    def _update_profile_group(self, oMenu, sType, fCallback, sProfile):
        """Update the profile selection menu"""
        oConfig = self._oMainWindow.config_file
        oGroup = oMenu.get_children()[0]

        aProfiles = [(sKey, oConfig.get_profile_option(sType, sKey, "name"))
            for sKey in oConfig.profiles(sType)]
        aProfiles.sort(key=lambda tProfile: tProfile[1])

        if sProfile is None or sProfile == 'Default':
            oGroup.set_active(True)

        # clear out existing radio items
        for oRadio in oGroup.get_group():
            if oRadio is not oGroup:
                oRadio.set_group(None)
                oMenu.remove(oRadio)

        for sKey, sName in aProfiles:
            oRadio = gtk.RadioMenuItem(oGroup, sName)
            oRadio.connect("toggled", fCallback, sKey)
            if sKey == sProfile:
                oRadio.set_active(True)
            oMenu.append(oRadio)
            oRadio.show()


class CardListMenu(FilteredViewMenu):
    # pylint: disable-msg=R0904
    # R0904 - gtk.Widget, so many public methods
    """Base class for Card List Menus

       Adds some common methods for dealing with the card lists -
       copying selections, etc.
       """

    def copy_selection(self, _oWidget):
        """Copy the current selection to the application clipboard."""
        self._oController.view.copy_selection()

    def create_analyze_menu(self):
        """Create the Analyze Menu, to be filled in by plugins"""
        self.create_submenu(self, "Analy_ze")



# ProgressDialog.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Dialog with progress meter
# Copyright 2007 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""classes needed for the progress dialog"""

import gtk
# pylint: disable-msg=W0402
# we need data from the string module
import string
# pylint: enable-msg=W0402
from logging import Handler


class SutekhLogHandler(Handler, object):
    """Base class for loggers to talk to the dialog"""
    # We explicitly inherit from object, since Handler is a classic class
    def __init__(self):
        super(SutekhLogHandler, self).__init__()
        self.oDialog = None

    def set_dialog(self, oDialog):
        """point to the progress dialog"""
        self.oDialog = oDialog

    def emit(self, _oRecord):
        """Default emit handler"""
        pass


class SutekhHTMLLogHandler(SutekhLogHandler):
    """Logging class for cardlist and rulings parser.

       Converts messages of the form 'Card: X' into an approximate
       progress measure
       """
    def __init__(self):
        super(SutekhHTMLLogHandler, self).__init__()

    # Massage messages from HTMLParser into Dialog updates
    def emit(self, oRecord):
        """Massage message into progress value.

           Skip difficult cases (The X, non-ascii characters, etc.)
           """
        if self.oDialog is None:
            return  # No point
        sString = oRecord.getMessage()
        # We skip considering the 'The' cases, as that gets messy
        if sString.startswith('Card: The '):
            return
        sBase = sString[6:8]
        sStart = string.upper(sBase)
        # The cardlist considers case in ordering - AK before Aabat, etc.
        # This causes jumps, so we ignore these, as there aren't enough
        # to make this worthwhile (9 out 3009, Jan 2008 - NM)
        if sStart == sBase:
            return
        # Skip the difficult, non-ascii encoding cases
        if sStart[0] not in string.ascii_uppercase or \
                sStart[1] not in string.ascii_uppercase:
            return
        # This is now safe, if not a particularly good idea
        iPos = ord(sStart[0]) * 26 + ord(sStart[1]) - (ord('A') * 27)
        fBarPos = iPos / float(26 * 26)
        self.oDialog.update_bar(fBarPos)


class SutekhCountLogHandler(SutekhLogHandler):
    """LogHandler class for dealing with database upgrade messages.

       Each message (Card List, card set, etc). is taken as a step in the
       process.
       """
    def __init__(self):
        super(SutekhCountLogHandler, self).__init__()
        self.iCount = None
        self.fTot = None

    def set_total(self, iTot):
        """Set the total number of steps."""
        self.fTot = float(iTot)
        self.iCount = 0

    def emit(self, _oRecord):
        """Handle a emitted signal, updating the progress count."""
        if self.oDialog is None:
            return  # No point
        self.iCount += 1
        fBarPos = self.iCount / self.fTot
        self.oDialog.update_bar(fBarPos)


class ProgressDialog(gtk.Window):
    # pylint: disable-msg=R0904
    # gtk.Widget, so many public methods
    """Show a window with a single progress bar."""
    # This is not a proper dialog, since we don't want the blocking
    # behaviour of Dialog.run()
    def __init__(self):
        super(ProgressDialog, self).__init__()
        self.set_title('Progress')
        self.set_name('Sutekh.dialog')
        self.oProgressBar = gtk.ProgressBar()
        self.oProgressBar.set_orientation(gtk.PROGRESS_LEFT_TO_RIGHT)
        self.oProgressBar.set_text('% done')
        self.oProgressBar.set_fraction(0.0)
        self.oProgressBar.set_size_request(350, 50)
        self.set_default_size(400, 100)
        self.oVBox = gtk.VBox()
        self.oDescription = gtk.Label('Unknown')
        self.oVBox.pack_start(self.oDescription)
        # Centre the progress bar horizontally
        oAlign = gtk.Alignment(xalign=0.5, yalign=0.0)
        oAlign.add(self.oProgressBar)
        self.oVBox.pack_end(oAlign)
        self.add(self.oVBox)
        self.set_type_hint(gtk.gdk.WINDOW_TYPE_HINT_UTILITY)
        self.show_all()
        self.set_modal(True)

    def set_description(self, sDescription):
        """Change the description of a dialog"""
        self.oVBox.remove(self.oDescription)
        self.oDescription = gtk.Label(sDescription)
        self.oVBox.pack_start(self.oDescription)
        self.show_all()

    def reset(self):
        """Reset the progress bar to zero"""
        self.update_bar(0.0)

    def set_complete(self):
        """Set the progress bar as being complete."""
        self.update_bar(1.0)

    def update_bar(self, fStep):
        """Update the progress bar to the given fStep value."""
        # Progress bar doesn't like values < 0.0 or > 1.0
        if fStep > 1.0:
            self.oProgressBar.set_fraction(1.0)
            self.oProgressBar.set_text('100% complete')
        elif fStep < 0.0:
            self.oProgressBar.set_fraction(0.0)
            self.oProgressBar.set_text(' 0% complete')
        else:
            self.oProgressBar.set_fraction(fStep)
            self.oProgressBar.set_text('%2.0f%% complete' % (fStep * 100))
        # Since Sutekh is single threaded, and some IO bound
        # task has the gtk.main loop process, we fudge the
        # required behaviour to get the update drawn
        while gtk.events_pending():
            gtk.main_iteration()



# test_CSVParser.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Test reading a card set from an ELDB deck file"""

import unittest
from sutekh.tests.TestCore import SutekhTest
from sutekh.io.CSVParser import CSVParser
from sutekh.tests.io.test_WriteCSV import EXPECTED_1, EXPECTED_2, EXPECTED_3, \
        EXPECTED_4


# Needs to be a SutekhTestCase so the name mapping cache test works
class TestCSVParser(SutekhTest):
    """class for the CSV deck file parser tests"""

    def test_basic(self):
        """Run the input test."""

        # Check we can round trip from our own CSV files
        aTests = [
                (0, 2, 1, True, EXPECTED_1),
                (0, 2, 1, False, EXPECTED_2),
                (0, 1, None, True, EXPECTED_3),
                (0, 1, None, False, EXPECTED_4),
                ]

        for tTestInfo in aTests:
            oParser = CSVParser(tTestInfo[0], tTestInfo[1], tTestInfo[2],
                    tTestInfo[3])

            oHolder = self._make_holder_from_string(oParser, tTestInfo[4])

            aCards = oHolder.get_cards()

            self.assertEqual(len(aCards), 11)
            self.failUnless((".44 Magnum", 4) in aCards)
            self.failUnless(("Abebe", 1) in aCards)
            self.failUnless(("Alan Sovereign (Advanced)", 1) in aCards)
            self.failUnless(("Aaron's Feeding Razor", 1) in aCards)
            self.failUnless(('Inez "Nurse216" Villagrande', 1) in aCards)
            self.failUnless(('Scapelli, The Family "Mechanic"', 1) in aCards)
            self.failUnless(("The Siamese", 1) in aCards)
            self.failUnless(("Abbot", 2) in aCards)

if __name__ == "__main__":
    unittest.main()



# MultiPaneWindow.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Handle the multi pane UI for Sutkeh
# Copyright 2007, 2008 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - See COPYING for details

"""Main window for Sutekh."""

# pylint: disable-msg=C0302
# C0302 - the module is long, but keeping the everything together is the best
# option for now

import pygtk
pygtk.require('2.0')
import gtk
import logging
# pylint: disable-msg=E0611
# pylint doesn't see resource_stream here, for some reason
from pkg_resources import resource_stream, resource_exists
from itertools import chain
# pylint: enable-msg=E0611
from sqlobject import SQLObjectNotFound
from sutekh.core.SutekhObjectCache import SutekhObjectCache
from sutekh.core.SutekhObjects import PhysicalCardSet, flush_cache, \
        PhysicalCard, IAbstractCard
from sutekh.gui.BasicFrame import BasicFrame
from sutekh.gui.PhysicalCardFrame import PhysicalCardFrame
from sutekh.gui.CardTextFrame import CardTextFrame
from sutekh.gui.CardSetFrame import CardSetFrame
from sutekh.gui.AboutDialog import SutekhAboutDialog
from sutekh.gui.MainMenu import MainMenu
from sutekh.gui.MainToolbar import MainToolbar
from sutekh.gui.GuiCardLookup import GuiLookup
from sutekh.gui.GuiCardSetFunctions import break_existing_loops
from sutekh.gui.CardSetManagementFrame import CardSetManagementFrame
from sutekh.gui.PluginManager import PluginManager
from sutekh.gui.GuiDBManagement import refresh_ww_card_list
from sutekh.gui import SutekhIcon
from sutekh.gui.HTMLTextView import HTMLViewDialog
from sutekh.gui.IconManager import IconManager
from sutekh.gui.SutekhDialog import do_complaint_error_details, \
        do_exception_complaint, do_complaint


class MultiPaneWindow(gtk.Window):
    """Window that has a configurable number of panes."""
    # pylint: disable-msg=R0904, R0902
    # R0904 - gtk.Widget, so many public methods
    # R0902 - we need to keep a lot of state, so many instance attributes
    def __init__(self):
        super(MultiPaneWindow, self).__init__(gtk.WINDOW_TOPLEVEL)
        self.set_name("Sutekh")
        self.set_title("Sutekh")
        self.set_border_width(2)

        # make sure we can quit
        self.connect("destroy", self.action_quit)
        self.connect("delete-event", self.action_close)
        # We can shrink the window quite small
        self.set_size_request(100, 100)
        # But we start at a reasonable size
        self.set_default_size(800, 600)

        # Set Default Window Icon for all Windows
        gtk.window_set_default_icon(SutekhIcon.SUTEKH_ICON)

        # common current directory for file dialogs
        self._sWorkingDir = ''

        # This always increments when panes are added, so the number is unique
        # for each new frame
        self._iCount = 0

        # Need to keep track of open card sets globally
        self.aClosedFrames = []
        self.aOpenFrames = []

        self._sCardSelection = ''  # copy + paste selection
        self._aHPanes = []
        self._aPlugins = []
        self.__dMenus = {}

        self._oFocussed = None

        # CardText frame is special, and there is only ever one of it
        # but we will set it up later
        self._bCardTextShown = False
        self._oCardTextPane = None
        self._oPCSListPane = None
        self._oHelpDlg = None
        # Global icon manager
        self._oIconManager = None
        self._oBusyCursor = gtk.gdk.Cursor(gtk.gdk.WATCH)
        # Sutekh lookuo cache
        self.__oSutekhObjectCache = None

    # pylint: disable-msg=W0201
    # We define attributes here, since this is called after database checks
    def setup(self, oConfig):
        """After database checks are passed, setup what we need to display
           data from the database."""
        self._oConfig = oConfig
        self._oCardLookup = GuiLookup(self._oConfig)

        # Check database is correctly populated

        try:
            _oCard = IAbstractCard('Ossian')
        except SQLObjectNotFound:
            # Log error so verbose picks it up
            logging.warn('Ossian not found in the database')
            # Inform the user
            iResponse = do_complaint(
                    'Database is missing cards. Try import the cardlist now?',
                    gtk.MESSAGE_ERROR, gtk.BUTTONS_YES_NO, False)
            if iResponse == gtk.RESPONSE_YES:
                refresh_ww_card_list(self)

        # Load plugins
        self._oPluginManager = PluginManager()
        self._oPluginManager.load_plugins()
        for cPlugin in self._oPluginManager.get_card_list_plugins():
            # Find plugins that will work on the Main Window
            self._aPlugins.append(cPlugin(self, None,
                "MainWindow"))
            cPlugin.register_with_config(oConfig)

        # Re-validate config after adding plugin specs
        oValidationResults = oConfig.validate()
        aErrors = oConfig.validation_errors(oValidationResults)
        if aErrors:
            logging.warn('Configuration file validation errors:\n',
                    "\n".join(aErrors))
            do_complaint_error_details(
                    "The configuration file validation failed:",
                    "\n".join(aErrors))
        oConfig.sanitize()

        # Create object cache
        self.__oSutekhObjectCache = SutekhObjectCache()

        # Create global icon manager
        self._oIconManager = IconManager(oConfig)
        # Create card text pane
        self._oCardTextPane = None  # So we can call get_pane_ids
        self._oCardTextPane = CardTextFrame(self, self._oIconManager)

        self.__oMenu = MainMenu(self, oConfig)
        self.__oToolbar = MainToolbar(self)

        # Do setup - prompt for downloads, etc. if needed
        self._oIconManager.setup()
        # plugins as well
        for oPlugin in self._aPlugins:
            oPlugin.setup()

        self.oVBox = gtk.VBox(False, 1)
        self.oVBox.pack_start(self.__oMenu, False, False)
        self.oVBox.pack_start(self.__oToolbar, False, False)
        self.oVBox.show()
        self.add(self.oVBox)

        # Break any loops in the database
        break_existing_loops()

        self.show_all()
        self.restore_from_config()
        self.connect('key-press-event', self.key_press)

    # pylint: enable-msg=W0201

    # pylint: disable-msg=W0212
    # We allow access via these properties
    # Needed for Backup plugin
    cardLookup = property(fget=lambda self: self._oCardLookup,
        doc="Used if user instance is needed to identify card names.")
    # Needed for plugins
    plugin_manager = property(fget=lambda self: self._oPluginManager,
            doc="The plugin manager for the application")
    plugins = property(fget=lambda self: self._aPlugins,
            doc="Plugins enabled for the main window.")
    config_file = property(fget=lambda self: self._oConfig,
            doc="The config file")
    focussed_pane = property(fget=lambda self: self._oFocussed,
            doc="The currently focussed pane.")
    mainwindow = property(fget=lambda self: self,
            doc="Return reference to the main window")
    icon_manager = property(fget=lambda self: self._oIconManager,
            doc="Used to lookup icons for disciplines, clans, etc.")
    card_text_pane = property(fget=lambda self: self._oCardTextPane,
            doc="Return reference to the card text pane")

    # pylint: enable-msg=W0212

    def add_to_menu_list(self, sMenuFlag, oMenuActiveFunc):
        """Add a key to the list of menu items to manage."""
        if not self.__dMenus.has_key(sMenuFlag):
            self.__dMenus[sMenuFlag] = oMenuActiveFunc

    # working directory methods

    def get_working_dir(self):
        """Get the current working dir for file chooser widgets"""
        return self._sWorkingDir

    def set_working_dir(self, sNewDir):
        """Set the working dir to sNewDir."""
        self._sWorkingDir = sNewDir

    # Config file handling

    def restore_from_config(self):
        """Restore all the frame form the config file."""
        # pylint: disable-msg=R0912, R0914, W9967
        # R0912: Need to consider all these cases, so many branches
        # R0914: Consequently, many local variables

        # Clear out all existing frames (loop over copies of the list)
        for oFrame in chain(self.aOpenFrames[:], self.aClosedFrames[:]):
            self.remove_frame(oFrame, True)

        # Flag to delay reloading pcs card list until after everything else
        # has loaded
        bReloadPCS = False

        aClosedFrames = []

        iWidth, iHeight = self._oConfig.get_window_size()
        if iWidth > 0 and iHeight > 0:
            self.resize(iWidth, iHeight)
        # Reset the pane number count, since we're starting afresh
        self._iCount = 0
        dPaneMap = {}
        for _iNumber, sType, sName, sOldId, bVert, bClosed, iPos in \
                self._oConfig.open_frames():
            oNewFrame = self.add_pane(bVert, iPos)
            oRestored = None
            self.win_focus(None, None, oNewFrame)
            if sType == PhysicalCardSet.sqlmeta.table:
                oRestored = self.replace_with_physical_card_set(sName,
                        oNewFrame, False)
                bReloadPCS = True
            elif sType == 'Card Text':
                oRestored = self.replace_with_card_text(None)
            elif sType == PhysicalCard.sqlmeta.table:
                oRestored = self.replace_with_physical_card_list(None)
            elif sType == 'Card Set List':
                oRestored = self.replace_with_pcs_list(None)
            else:
                # See if one of the plugins claims this type
                for oPlugin in self._aPlugins:
                    oFrame = oPlugin.get_frame_from_config(sType)
                    if oFrame:
                        self.replace_frame(oNewFrame, oFrame)
                        oRestored = oFrame
            if bClosed and oRestored:
                # Hold off minimizing until all frames are added to avoid
                # corner cases around the 1st restored frame needing to be
                # minimised
                aClosedFrames.append(oRestored)
            if oRestored and sOldId:
                dPaneMap[sOldId] = oRestored.config_frame_id
        for oFrame in aClosedFrames:
            self.minimize_to_toolbar(oFrame)
        if dPaneMap:
            self._oConfig.update_pane_numbers(dPaneMap)
        if bReloadPCS:
            self.reload_pcs_list()
        if not self.aOpenFrames:
            # We always have at least one pane
            self.add_pane()

    # Pane manipulation functions

    def top_level_panes(self):
        """Return a list of top-level panes.

           This is just the contents of the MainWindow vbox minus
           the toolbar and menu.
           """
        return [x for x in self.oVBox.get_children() if
                    x not in (self.__oMenu, self.__oToolbar)]

    def find_pane_by_id(self, iId):
        """Return the gtk widget corresponding to the given pane name"""
        for oPane in chain(self.aOpenFrames, self.aClosedFrames):
            if oPane.pane_id == iId:
                return oPane
        return None

    def get_pane_ids(self):
        """Return a list of all relevant pane ids"""
        aIds = []
        if self._oCardTextPane:
            aIds.append(self._oCardTextPane.pane_id)
        aIds.extend([x.pane_id for x in self.aOpenFrames])
        aIds.extend([x.pane_id for x in self.aClosedFrames])
        return aIds

    def is_open_by_menu_name(self, sMenuFlag):
        """Check if a frame with the given menu name is open"""
        for oPane in chain(self.aOpenFrames, self.aClosedFrames):
            if oPane.get_menu_name() == sMenuFlag:
                return True
        return False

    def find_cs_pane_by_set_name(self, sSetName):
        """Find all the panes that correspond to the same card set"""
        aPanes = []
        for oPane in chain(self.aOpenFrames, self.aClosedFrames):
            if oPane.is_card_set(sSetName):
                aPanes.append(oPane)
        return aPanes

    def replace_with_physical_card_set(self, sName, oFrame, bDoReloadPCS=True,
            bStartEditable=False):
        """Replace the pane oFrame with the physical card set sName"""
        if oFrame:
            # pylint: disable-msg=W0704
            # not doing anything for errors right now
            try:
                oPane = CardSetFrame(self, sName, bStartEditable)
                self.replace_frame(oFrame, oPane, bDoReloadPCS)
                return oPane
            except RuntimeError, oExp:
                do_exception_complaint("Unable to open Card Set %s\n"
                        "Error: %s" % (sName, oExp))
        return None

    def add_new_physical_card_set(self, sName, bStartEditable=False):
        """Create a new pane and replace with the PCS sName"""
        oFrame = self.add_pane_end()
        return self.replace_with_physical_card_set(sName, oFrame,
                bStartEditable=bStartEditable)

    def replace_with_pcs_list(self, _oWidget, oOldFrame=None):
        """Replace the focussed or given pane with the physical card set
           list."""
        if self.is_open_by_menu_name("Card Set List"):
            return None
        if oOldFrame is None:
            oOldFrame = self._oFocussed
        oPane = CardSetManagementFrame(self)
        self.replace_frame(oOldFrame, oPane)
        self._oPCSListPane = oPane
        return oPane

    def add_new_pcs_list(self, oMenuWidget):
        """Add a new pane and replace it with the physical card set list."""
        oNewPane = self.add_pane_end()
        return self.replace_with_pcs_list(oMenuWidget, oNewPane)

    def replace_with_physical_card_list(self, _oWidget, oOldFrame=None):
        """Replace the currently focussed or given pane with the physical
           card list."""
        if self.is_open_by_menu_name("White Wolf Card List"):
            return None
        if oOldFrame is None:
            oOldFrame = self._oFocussed
        oPane = PhysicalCardFrame(self)
        self.replace_frame(oOldFrame, oPane)
        return oPane

    def add_new_physical_card_list(self, oMenuWidget):
        """Add a new pane and replace it with the physical card list."""
        oNewPane = self.add_pane_end()
        return self.replace_with_physical_card_list(oMenuWidget, oNewPane)

    def replace_with_card_text(self, _oWidget, oOldFrame=None):
        """Replace the current or given pane with the card set pane."""
        if self.is_open_by_menu_name("Card Text"):
            return None
        if oOldFrame is None:
            oOldFrame = self._oFocussed
        self.replace_frame(oOldFrame, self._oCardTextPane)
        return self._oCardTextPane

    def add_new_card_text(self, oMenuWidget):
        """Add a new pane and replace it with the card set pane."""
        oNewPane = self.add_pane_end()
        return self.replace_with_card_text(oMenuWidget, oNewPane)

    # state manipulation
    def reload_pcs_list(self):
        """Reload the list of physical card sets."""
        if self._oPCSListPane is not None:
            self._oPCSListPane.reload()

    def reload_all(self):
        """Reload all frames. Useful for major DB changes"""
        for oPane in chain(self.aOpenFrames, self.aClosedFrames):
            oPane.reload()

    def do_all_queued_reloads(self):
        """Do any deferred reloads from the database signal handlers."""
        for oPane in chain(self.aOpenFrames, self.aClosedFrames):
            oPane.do_queued_reload()

    def update_to_new_db(self):
        """Resync panes against the database.

           Needed because ids aren't kept across re-reading the WW
           cardlist, since card sets with children are always created
           before there children are added.
           """
        # Flush the caches, so we don't hit stale lookups
        flush_cache()
        # Reset the lookup cache holder
        self.__oSutekhObjectCache = SutekhObjectCache()
        # We may close frames here, so loop over copies of the list
        for oPane in chain(self.aOpenFrames[:], self.aClosedFrames[:]):
            oPane.update_to_new_db()

    def clear_cache(self):
        """Remove the cached set of objects, for card list reloads, etc."""
        del self.__oSutekhObjectCache

    def get_editable_panes(self):
        """Get a list of panes, which are currently editable.

           Used by the WW import code, and backup restores to correct
           for the zero card list state setting the cardlist's to editable.
           """
        aEditable = []
        for oPane in chain(self.aOpenFrames, self.aClosedFrames):
            if hasattr(oPane.view, 'toggle_editable'):
                if oPane.view.get_model().bEditable:
                    aEditable.append(oPane)
        return aEditable

    def restore_editable_panes(self, aEditable):
        """Restore the editable state so only the panes in aEditable are
           editable."""
        for oPane in chain(self.aOpenFrames, self.aClosedFrames):
            if hasattr(oPane.view, 'toggle_editable'):
                if oPane in aEditable:
                    oPane.view.toggle_editable(True)
                else:
                    oPane.view.toggle_editable(False)

    def set_card_text(self, oCard):
        """Update the card text frame to the currently selected card."""
        if oCard:
            # Skip if None
            self._oCardTextPane.view.set_card_text(oCard)

    def reset_menu(self):
        """Ensure menu state is correct."""
        for fSetSensitiveFunc in self.__dMenus.values():
            fSetSensitiveFunc(True)
        for oPane in chain(self.aOpenFrames, self.aClosedFrames):
            sMenuName = oPane.get_menu_name()
            if sMenuName in self.__dMenus:
                self.__dMenus[sMenuName](False)
        # Disable the menu accelerators for unfocussed panes
        for oFrame in chain(self.aOpenFrames, self.aClosedFrames):
            if self._oFocussed != oFrame and hasattr(oFrame, 'menu') and \
                    hasattr(oFrame.menu, 'remove_accels'):
                oFrame.menu.remove_accels()
        # Handle focussed pane stuff
        if self._oFocussed:
            # Can always split horizontally
            self.__oMenu.set_split_horizontal_active(True)
            self.__oMenu.replace_pane_set_sensitive(True)
            # But we can't split vertically more than once
            if isinstance(self._oFocussed.get_parent(), gtk.VPaned):
                self.__oMenu.set_split_vertical_active(False)
            else:
                self.__oMenu.set_split_vertical_active(True)
            # can't delete the last pane
            if len(self.aOpenFrames) > 1:
                self.__oMenu.del_pane_set_sensitive(True)
            else:
                self.__oMenu.del_pane_set_sensitive(False)
            # Ensure accelerators are active
            if hasattr(self._oFocussed, 'menu') and \
                    hasattr(self._oFocussed.menu, 'activate_accels'):
                self._oFocussed.menu.activate_accels()
        else:
            # Can't split when no pane chosen
            self.__oMenu.set_split_vertical_active(False)
            self.__oMenu.set_split_horizontal_active(False)
            # Can't delete either
            self.__oMenu.del_pane_set_sensitive(False)
            self.__oMenu.replace_pane_set_sensitive(False)

    def set_selection_text(self, sText):
        """Set the current selection text for copy+paste."""
        self._sCardSelection = sText

    def get_selection_text(self):
        """Get the current selection for copy+paste."""
        return self._sCardSelection

    # startup, exit and other misc functions

    # pylint: disable-msg=R0201
    # making this a function would not be convient
    def run(self):
        """gtk entry point"""
        gtk.main()

    def action_quit(self, _oWidget):
        """Exit the app"""
        # ensure we cleanup all signals (needed for tests)
        for oFrame in chain(self.aOpenFrames, self.aClosedFrames):
            oFrame.cleanup()
        for oPlugin in self._aPlugins:
            oPlugin.cleanup()
        # don't hold references to plugins here either
        self._aPlugins = []
        # Don't call gtk.main_quit when the main loop isn't running (true in
        # the tests)
        if gtk.main_level() > 0:
            gtk.main_quit()

    def show_about_dialog(self, _oWidget):
        """Display the about dialog"""
        oDlg = SutekhAboutDialog()
        oDlg.run()
        oDlg.destroy()

    def _link_resource(self, sLocalUrl):
        """Return a file-like object which sLocalUrl can be read from."""
        sResource = '/docs/html/%s' % sLocalUrl
        if resource_exists('sutekh', sResource):
            return resource_stream('sutekh', sResource)
        else:
            raise ValueError("Unknown resource %s" % sLocalUrl)

    # pylint: enable-msg=R0201

    def win_focus(self, _oWidget, _oEvent, oFrame):
        """Respond to focus change events.

           Keep track of focussed pane, update menus and handle highlighting.
           """
        if self._oFocussed is not None:
            self._oFocussed.set_unfocussed_title()
        self._oFocussed = oFrame
        if oFrame:
            # oFrame can be None when win_focus is called directly
            self._oFocussed.set_focussed_title()
            self._oFocussed.view.grab_focus()
        self.reset_menu()

    def key_press(self, _oWidget, oEvent):
        """Move to the next frame on Tab"""
        if oEvent.keyval == gtk.gdk.keyval_from_name('Tab') or \
                oEvent.keyval == gtk.gdk.keyval_from_name('KP_Tab'):
            self.move_to_frame(+1)
            return True
        elif oEvent.keyval == gtk.gdk.keyval_from_name('ISO_Left_Tab'):
            # Shift & tab should always be this, AFAICT
            self.move_to_frame(-1)
            return True
        return False  # propogate event

    def action_close(self, _oWidget, _oEvent):
        """Close the app (menu or window manager) and save the settings"""
        if self._oConfig.get_save_on_exit():
            self.save_frames()
        if self._oConfig.get_save_window_size():
            self.save_window_size()
        # cue app exit
        self.destroy()

    def show_tutorial(self, _oMenuWidget, oHelpLast):
        """Show the HTML Tutorial"""
        fTutorial = self._link_resource('Tutorial.html')
        oHelpLast.set_sensitive(True)
        self._do_html_dialog(fTutorial)

    def show_manual(self, _oMenuWidget, oHelpLast):
        """Show the HTML Manual"""
        fManual = self._link_resource('Manual.html')
        oHelpLast.set_sensitive(True)
        self._do_html_dialog(fManual)

    def show_last_help(self, _oMenuWidget):
        """Reshow the help dialog with the last shown page"""
        if self._oHelpDlg is not None:
            self._oHelpDlg.show()

    # _oWidget needed so this can be also be called from the menu
    def add_pane_end(self, _oWidget=None):
        """Add a pane to the right edge of the window.

           Used for the add pane menu items and double clicking on card set
           names.
           """
        if not self.aOpenFrames:
            oPane = None
        else:
            # Descend the right child of the panes, until we get a
            # non-paned item
            oPane = self.top_level_panes()[0]
            while isinstance(oPane, (gtk.HPaned, gtk.VPaned)):
                oPane = oPane.get_child2()
        oOldFocus = self._oFocussed
        self.win_focus(None, None, oPane)
        oNewPane = self.add_pane()
        self.win_focus(None, None, oOldFocus)
        return oNewPane

    def menu_remove_frame(self, _oMenuWidget):
        """Handle the remove pane request from the menu"""
        # Closing the pane via the menu, so remove any existing pane profiles
        self._oConfig.clear_frame_profile(self._oFocussed.config_frame_id)
        self.remove_frame(self._oFocussed)

    def _do_html_dialog(self, fInput):
        """Popup and run HTML Dialog widget"""
        if self._oHelpDlg is None:
            self._oHelpDlg = HTMLViewDialog(self, fInput, self._link_resource)
        else:
            self._oHelpDlg.show_page(fInput)
        self._oHelpDlg.show()

    # window setup saving functions

    def save_window_size(self):
        """Write the current window size to the config file"""
        self._oConfig.set_window_size(self.get_size())

    def save_frames(self):
        # pylint:disable-msg=R0912
        # lots of decisions, so many branches
        """save the current frame layout"""
        self._oConfig.clear_open_frames()

        # pylint: disable-msg=R0913
        # we need all these arguments
        def save_pane_to_config(iNum, oPane, oConfig, bVert, bClosed, iPos):
            """Save a pane to the config file, dealing with card set panes
               correctly"""
            if oPane.type != PhysicalCardSet.sqlmeta.table:
                oConfig.add_frame(iNum, oPane.type, oPane.name, bVert,
                        bClosed, iPos, oPane.config_frame_id)
            else:
                oConfig.add_frame(iNum, oPane.type, oPane.cardset_name, bVert,
                        bClosed, iPos, oPane.config_frame_id)

        def save_children(oPane, oConfig, bVert, iNum, iPos):
            """Walk the tree of HPanes + VPanes, and save their info."""
            # oPane.get_position() gives us the position relative to the paned
            # we're contained in. However, when we recreate the layout, we
            # don't split panes in the same order, hence the fancy
            # score-keeping work to convert the obtained positions to those
            # needed for restoring
            if isinstance(oPane, gtk.HPaned):
                oChild1 = oPane.get_child1()
                oChild2 = oPane.get_child2()
                iNum, iChildPos = save_children(oChild1, oConfig, False,
                        iNum, iPos)
                iMyPos = oPane.get_position()
                if isinstance(oChild1, gtk.HPaned):
                    iMyPos = iMyPos - iChildPos
                    iChildPos = oPane.get_position()
                else:
                    iChildPos = iMyPos
                if iMyPos < 1:
                    # Setting pos to < 1 doesn't do what we want
                    iMyPos = 1
                iNum, iChild2Pos = save_children(oChild2, oConfig, False,
                        iNum, iMyPos)
                if isinstance(oChild2, gtk.HPaned):
                    iChildPos += iChild2Pos
            elif isinstance(oPane, gtk.VPaned):
                oChild1 = oPane.get_child1()
                oChild2 = oPane.get_child2()
                iMyPos = oPane.get_position()
                iNum, _iTemp = save_children(oChild1, oConfig, False, iNum,
                        iPos)
                iNum, _iTemp = save_children(oChild2, oConfig, True, iNum,
                        iMyPos)
                iChildPos = iPos
            else:
                save_pane_to_config(iNum, oPane, oConfig, bVert, False, iPos)
                iNum += 1
                iChildPos = iPos
            return iNum, iChildPos

        iNum = 1
        aTopLevelPanes = self.top_level_panes()
        if aTopLevelPanes:
            iNum, _iPos = save_children(aTopLevelPanes[0], self._oConfig,
                    False, 1, -1)
        for oPane in self.aClosedFrames:
            save_pane_to_config(iNum, oPane, self._oConfig, False, True, -1)
            iNum += 1

    # frame management functions

    def move_to_frame(self, iDir):
        """Move the focus to the previous frame"""
        if iDir > 0:
            oNewFrame = self._find_next_frame(self._oFocussed)
        else:
            oNewFrame = self._find_prev_frame(self._oFocussed)
        if oNewFrame:
            self.win_focus(None, None, oNewFrame)

    def _find_prev_frame(self, oFrame):
        """Move the focus to the next frame in the window"""
        if not oFrame:
            if not self.aOpenFrames:
                # No frame, so nothing to focus
                return
            # get first child, and descent that
            oChild = self.top_level_panes()[0]
            oParent = self.oVBox
        else:
            oParent = oFrame.get_parent()
            if isinstance(oParent, gtk.VPaned):
                if oParent.get_child2() == oFrame:
                    # Need to go to the top frame
                    return oParent.get_child1()
                oChild = oParent
            else:
                oChild = oFrame
            # Need to move to the 'neighbouring' HPane
            oParent = oChild.get_parent()
            while isinstance(oParent, gtk.HPaned) and \
                        oChild == oParent.get_child1():
                # As long as we're the left-most branch, we ascend
                oChild = oParent
                oParent = oChild.get_parent()
        # prepare to descend to the correct frame
        if isinstance(oParent, gtk.HPaned) and oChild == oParent.get_child2():
            # We need to start going down the left-hand tree
            oChild = oParent.get_child1()
        while isinstance(oChild, gtk.HPaned):
            # we descend down the right branch until we reach a frame
            oChild = oChild.get_child2()
        if isinstance(oChild, gtk.VPaned):
            # go to the bottom frame in this case
            oChild = oChild.get_child2()
        return oChild

    def _find_next_frame(self, oFrame):
        """Move the focus to the next frame in the window"""
        if not oFrame:
            if not self.aOpenFrames:
                # No frame, so nothing to focus
                return
            # get first child, and descent that
            oChild = self.top_level_panes()[0]
            oParent = self.oVBox
        else:
            oParent = oFrame.get_parent()
            if isinstance(oParent, gtk.VPaned):
                if oParent.get_child1() == oFrame:
                    # Need to go to the bottom frame
                    return oParent.get_child2()
                oChild = oParent
            else:
                oChild = oFrame
            # Need to move to the 'neighbouring' HPane
            oParent = oChild.get_parent()
            while isinstance(oParent, gtk.HPaned) and \
                        oChild == oParent.get_child2():
                # As long as we're the right-most branch, we ascend
                oChild = oParent
                oParent = oChild.get_parent()
        # prepare to descend to the correct frame
        if isinstance(oParent, gtk.HPaned) and oChild == oParent.get_child1():
            # We need to start going down the right-hand tree
            oChild = oParent.get_child2()
        while isinstance(oChild, gtk.HPaned):
            # we descend down the left branch until we reach a frame
            oChild = oChild.get_child1()
        if isinstance(oChild, gtk.VPaned):
            # go to the top frame in this case
            oChild = oChild.get_child1()
        return oChild

    def replace_frame(self, oOldFrame, oNewFrame, bDoReloadPCS=True):
        """Replace oOldFrame with oNewFrame + update menus"""
        oNewFrame.show()
        oNewFrame.set_focus_handler(self.win_focus)
        oParent = oOldFrame.get_parent()
        oParent.remove(oOldFrame)
        oParent.add(oNewFrame)
        self.aOpenFrames.remove(oOldFrame)
        self.aOpenFrames.append(oNewFrame)
        if oNewFrame in self.aClosedFrames:
            self.aClosedFrames.remove(oNewFrame)
        if self._oFocussed == oOldFrame:
            self.win_focus(None, None, None)
        # kill any listeners, etc.
        oOldFrame.cleanup()
        self.reset_menu()
        # Open card lists may have changed because of the frame we've
        # kicked out
        if bDoReloadPCS:
            self.reload_pcs_list()

    def swap_frames(self, oFrame1, oFrame2):
        """swap two frames - used by drag-n-drop code"""
        if oFrame1 != oFrame2:
            oParent1 = oFrame1.get_parent()
            oParent2 = oFrame2.get_parent()
            if oParent1 == oParent2:
                # Special logic to ensure that we get them swapped
                if oFrame1 == oParent1.get_child1():
                    oParent1.remove(oFrame1)
                    oParent1.remove(oFrame2)
                    oParent1.add1(oFrame2)
                    oParent1.add2(oFrame1)
                else:
                    oParent1.remove(oFrame1)
                    oParent1.remove(oFrame2)
                    oParent1.add1(oFrame1)
                    oParent1.add2(oFrame2)
            else:
                oParent2.remove(oFrame2)
                oParent1.remove(oFrame1)
                oParent2.add(oFrame1)
                oParent1.add(oFrame2)
            oParent1.show()
            oParent2.show()
            self.reset_menu()

    def get_current_pane(self):
        """Get the parent HPane of the focussed pane.

           If there's no Focussed Pane, return the last added pane,
           which does the right thing for restore_from_config.
           """
        if self._oFocussed:
            oParent = self._oFocussed.get_parent()
            if isinstance(oParent, gtk.VPaned):
                # Get the HPane this belongs to
                oParent = oParent.get_parent()
            return oParent
        else:
            # Return the last added pane, for automatic adds
            return self._aHPanes[-1]

    # Utility functions for add_pane
    def _get_pane_to_replace(self):
        """Get the pane to be replaced in add_pane

           Return the pane and the part to replace"""
        oParent = self.get_current_pane()
        # Must be a hpane, by construction
        if self._oFocussed:
            oPart1 = self._oFocussed
            if isinstance(oPart1.get_parent(), gtk.VPaned):
                # Veritical pane, so we need to use the pane, not
                # the Frame
                oPart1 = oPart1.get_parent()
        else:
            # Replace right child of last added pane when no
            # obvious option
            oPart1 = oParent.get_child2()
        return oPart1, oParent

    def _do_replace_pane(self, oParent, oPart1, oNewPane, iConfigPos):
        """Handle the nitty gritty of replacing a pane in the window"""
        if oPart1 == oParent.get_child1():
            oParent.remove(oPart1)
            oParent.add1(oNewPane)
            # Going to the left of the current pane,
            iPos = oParent.get_position() / 2
        else:
            oParent.remove(oPart1)
            oParent.add2(oNewPane)
            oCur = oParent.get_allocation()
            if oCur.width == 1 and oParent.get_position() > 1 and \
                    iConfigPos == -1:
                # we are in early startup, so we can move the
                # Parent as well
                # We want to split ourselves into equally sized
                # sections
                oCur = self.oVBox.get_allocation()
                iPos = oCur.width / (len(self._aHPanes) + 2)
                self.set_pos_for_all_hpanes(iPos)
            else:
                iPos = (oCur.width - oParent.get_position()) / 2
        return iPos

    def add_pane(self, bVertical=False, iConfigPos=-1):
        """Add a blank frame to the window.

           bVertical -> set True to split the current pane vertically
           iConfigPos -> Layout parameter for restoring positions.
           if iConfigPos == -1, the currently focussed frame is
           halved in size.
           """
        # this is a complex function (arguably too complex). Simplifying
        # it is non trivial
        oWidget = BasicFrame(self)
        oWidget.add_parts()
        oWidget.set_focus_handler(self.win_focus)
        self._iCount += 1
        sKey = 'Blank Pane:' + str(self._iCount)
        oWidget.set_title(sKey)
        self.aOpenFrames.append(oWidget)
        if len(self.aOpenFrames) == 1:
            # We have a blank space to fill, so just plonk in the widget
            self.oVBox.pack_start(oWidget)
        else:
            # We already have a widget, so we add a pane
            if bVertical:
                oNewPane = gtk.VPaned()
                oCurAlloc = self.oVBox.get_allocation()
                oMenuAlloc = self.__oMenu.get_allocation()
                iPos = (oCurAlloc.height - oMenuAlloc.height) / 2
            else:
                oNewPane = gtk.HPaned()
            if len(self._aHPanes) > 0:
                # We pop out the current frame, and plonk it in
                # the new pane - we add the new widget to the other
                # part
                oPart1, oParent = self._get_pane_to_replace()
                iTPos = self._do_replace_pane(oParent, oPart1, oNewPane,
                        iConfigPos)
                if not bVertical:
                    iPos = iTPos
            else:
                # This is the first HPane, or the 1st VPane we add, so
                # vbox has 2 children
                # The menu, and the one we want
                oPart1 = self.top_level_panes()[0]
                oParent = self.oVBox
                # Just split the window
                oCur = oParent.get_allocation()
                if not bVertical:
                    iPos = oCur.width / 2
                oParent.remove(oPart1)
                oParent.pack_start(oNewPane)
            oNewPane.add1(oPart1)
            oNewPane.add2(oWidget)
            oPart1.show()
            oNewPane.show()
            oParent.show()
            if not bVertical:
                self._aHPanes.append(oNewPane)
            if iConfigPos == -1:
                oNewPane.set_position(iPos)
            else:
                oNewPane.set_position(iConfigPos)
        self.oVBox.show()
        self.reset_menu()
        # Reset frame focussed state
        if self._oFocussed:
            self.win_focus(None, None, self._oFocussed)
        return oWidget

    def remove_frame(self, oFrame, bForceZero=False, bClose=False):
        """Remove the Frame oFrame from the window.

           bForceZero is used to force the removal of the last widget in
           the window.

           bClose is used to add the frame to the list of closed frames
           rather than destroying it and cleaning it up.
           """
        if oFrame is not None and oFrame in self.aOpenFrames:
            oNeighbour = self._find_prev_frame(oFrame)
            if len(self.aOpenFrames) == 1:
                if not bForceZero:
                    # Break out, as we do nothing
                    return
                # Removing last widget, so just clear the vbox
                oWidget = self.top_level_panes()[0]
                self.oVBox.remove(oWidget)
            elif isinstance(oFrame.get_parent(), gtk.VPaned):
                # Removing a vertical frame, keep the correct child
                oParent = oFrame.get_parent()
                oKept = [x for x in oParent.get_children() if x != oFrame][0]
                oHPane = oParent.get_parent()
                oHPane.remove(oParent)
                oParent.remove(oFrame)
                oParent.remove(oKept)
                oHPane.add(oKept)
            elif len(self._aHPanes) == 1:
                # Removing from the only pane, so keep the other pane
                oThisPane = self._aHPanes[0]  # Only pane
                self._aHPanes.remove(oThisPane)
                oKept = [x for x in oThisPane.get_children() if x != oFrame][0]
                # clear out pane
                oThisPane.remove(oFrame)
                oThisPane.remove(oKept)
                self.oVBox.remove(oThisPane)
                self.oVBox.pack_start(oKept)
            else:
                oFocussedPane = [x for x in self._aHPanes if oFrame in
                        x.get_children()][0]
                oKept = [x for x in oFocussedPane.get_children() if
                        x != oFrame][0]
                oParent = oFocussedPane.get_parent()
                oParent.remove(oFocussedPane)
                oFocussedPane.remove(oKept)
                oParent.add(oKept)
                # Housekeeping
                oFocussedPane.remove(oFrame)
                self._aHPanes.remove(oFocussedPane)
            self.oVBox.show()
            # Remove from list of open panes
            self.aOpenFrames.remove(oFrame)
            # Any cleanup events we need?
            if bClose:
                self.aClosedFrames.append(oFrame)
            else:
                oFrame.cleanup()
            if oFrame == self._oFocussed:
                if oNeighbour != oFrame:
                    # Removing the focussed frame, so we move
                    self.win_focus(None, None, oNeighbour)
                else:
                    # Can't do better than this, really
                    self.win_focus(None, None, None)
            self.reset_menu()
        elif oFrame in self.aClosedFrames and not bClose:
            self.aClosedFrames.remove(oFrame)
            # Need to remove button from the toolbar
            self.__oToolbar.remove_frame_button(oFrame.title)
            oFrame.cleanup()

    def set_all_panes_equal(self):
        """Evenly distribute the space between all the frames."""
        oCurAlloc = self.oVBox.get_allocation()
        iNewPos = oCurAlloc.width / (len(self._aHPanes) + 1)
        self.set_pos_for_all_hpanes(iNewPos)

    def set_pos_for_all_hpanes(self, iNewPos):
        """Set all the panes to the same Position value"""
        oCurAlloc = self.oVBox.get_allocation()
        oMenuAlloc = self.__oMenu.get_allocation()
        iVertPos = (oCurAlloc.height - oMenuAlloc.height) / 2

        def set_pos_children(oPane, iPos, iVertPos):
            """Walk the tree in display order, setting positions accordingly"""
            oChild1 = oPane.get_child1()
            oChild2 = oPane.get_child2()
            if isinstance(oChild1, gtk.HPaned):
                iMyPos = iPos + set_pos_children(oChild1, iPos, iVertPos)
            else:
                if isinstance(oChild1, gtk.VPaned):
                    oChild1.set_position(iVertPos)
                iMyPos = iPos
            oPane.set_position(iMyPos)
            if isinstance(oChild2, gtk.HPaned):
                iMyPos += set_pos_children(oChild2, iPos, iVertPos)
            elif isinstance(oChild2, gtk.VPaned):
                oChild2.set_position(iVertPos)
            return iMyPos

        aTopLevelPanes = self.top_level_panes()
        for oPane in aTopLevelPanes:
            # Should only be 1 here
            if isinstance(oPane, gtk.HPaned):
                set_pos_children(oPane, iNewPos, iVertPos)
            elif isinstance(oPane, gtk.VPaned):
                # Do something sensible for single VPane case
                oPane.set_position(iVertPos)

    def set_busy_cursor(self):
        """Set the window cursor to indicate busy status"""
        # This needs to be on the top level widget, so has to be here
        # pylint: disable-msg=E1101
        # gtk properties confuse pylint here
        # Safe-guard for if we're called while shutting down
        if self.window:
            self.window.set_cursor(self._oBusyCursor)
            # Since we're about to do a bunch of CPU grinding, tell gtk
            # to redraw now
            gtk.gdk.flush()

    def restore_cursor(self):
        """Restore the ordinary cursor"""
        # pylint: disable-msg=E1101
        # gtk properties confuse pylint here
        # Safe-guard for if we're called while shutting down
        if self.window:
            self.window.set_cursor(None)

    def minimize_to_toolbar(self, oFrame):
        """Minimize the given frame to the main menu"""
        # don't allow last frame to be minimized
        if len(self.aOpenFrames) == 1:
            return
        self.__oToolbar.frame_to_toolbar(oFrame, oFrame.title)
        self.remove_frame(oFrame, bClose=True)

    def restore_from_toolbar(self, oFrame):
        """Restore the given frame from the toolbar"""
        oNewPane = self.add_pane_end()
        self.replace_frame(oNewPane, oFrame)
        self.win_focus(None, None, oFrame)



# PreferenceTable.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Table widget for editing a list of options.
# Copyright 2010 Simon Cross <hodgestar+sutekh@gmail.com>
# GPL - see COPYING for details

"""Widget that constructs an interface to manipulate the profile options.

   Used by the various Profile editor dialogs."""

import gtk


class PreferenceTable(gtk.Table):
    """A widget for editing a list of options."""
    # pylint: disable-msg=R0904, R0902
    # R0904 - gtk.Widget, so many public methods

    COLUMNS = 3
    KEY_COL, ENTRY_COL, INHERIT_COL = range(COLUMNS)

    def __init__(self, aOptions, oValidator):
        self._aOptions = []
        for sKey, sConfigSpec, _bInherit in aOptions:
            self._aOptions.append(
                PreferenceOption(sKey, sConfigSpec, oValidator))

        self._aOptions.sort(
            key=lambda oOpt: oOpt.sKey != "name" and oOpt.sKey or "")

        super(PreferenceTable, self).__init__(
            rows=len(self._aOptions), columns=self.COLUMNS)
        self.set_col_spacings(5)
        self.set_row_spacings(5)

        dAttachOpts = {
            "xoptions": gtk.FILL,
            "yoptions": 0,
        }
        for iRow, oOpt in enumerate(self._aOptions):
            # pylint: disable-msg=W0142
            # *args magic OK here
            self.attach(oOpt.oLabel, self.KEY_COL, self.KEY_COL + 1,
                iRow, iRow + 1, **dAttachOpts)
            self.attach(oOpt.oEntry, self.ENTRY_COL, self.ENTRY_COL + 1,
                iRow, iRow + 1, **dAttachOpts)
            self.attach(oOpt.oInherit, self.INHERIT_COL, self.INHERIT_COL + 1,
                iRow, iRow + 1, **dAttachOpts)

    def update_values(self, dNewValues, dInherit, dEditable, dInheritedValues):
        """Update the option values.

           dNewValues is an sKey -> sValue mapping.
           dInherit is an sKey -> bInheritable mapping.
           dEditable is an sKey -> bEditable mapping.
           """
        for oOpt in self._aOptions:
            oInheritedValue = dInheritedValues.get(oOpt.sKey)
            oOpt.set_inheritable(dInherit.get(oOpt.sKey, True))
            oOpt.set_editable(dEditable.get(oOpt.sKey, True))
            oOpt.set_value(dNewValues.get(oOpt.sKey), oInheritedValue)

    def get_values(self):
        """Return a dictionary of option values."""
        return dict((oOpt.sKey, oOpt.get_value()) for oOpt in self._aOptions)


class PreferenceOption(object):
    """An option for a preference table."""

    def __init__(self, sKey, sConfigSpec, oValidator):
        self.sKey = sKey
        self.oSpec = parse_spec(sConfigSpec, oValidator)
        self.oEntry = self.oSpec.oEntry
        self.oLabel = gtk.Label(sKey.capitalize())
        self.oLabel.set_alignment(0, 0)
        self.bInheritable = True
        self.oInherit = gtk.CheckButton("use default")
        self.oInherit.connect("toggled", self._inherit_toggled)
        self.bEditable = True

    def set_inheritable(self, bInheritable):
        """Set whether the values are inherited from the default profile"""
        self.bInheritable = bInheritable
        if bInheritable:
            self.oInherit.set_sensitive(True)
            self.oInherit.show()
        else:
            self.oInherit.set_sensitive(False)
            self.oInherit.hide()

    def set_editable(self, bEditable):
        """Mark the option as editable"""
        self.bEditable = bEditable
        self.oEntry.set_sensitive(bEditable)

    def set_value(self, oValue, oInheritedValue):
        """Update the value of the option."""
        if oValue is None and self.bInheritable:
            self.oEntry.set_sensitive(False)
            self.oInherit.set_active(True)
        else:
            self.oEntry.set_sensitive(self.bEditable)
            self.oInherit.set_active(False)
            self.oSpec.set_value(oValue)
        if hasattr(self.oInherit, 'set_tooltip_markup'):
            sToolTip = "Inherited value:\n%s" % \
                self.oSpec.format_value(oInheritedValue)
            self.oInherit.set_tooltip_markup(sToolTip)

    def get_value(self):
        """Return the value of the option."""
        if self.oInherit.get_active():
            return None
        else:
            return self.oSpec.get_value()

    def _inherit_toggled(self, _oWidget):
        """Update the state after the inherit widget is toggled."""
        if self.oInherit.get_active():
            self.oEntry.set_sensitive(False)
        else:
            self.oEntry.set_sensitive(self.bEditable)


class BaseParsedSpec(object):
    """Object holding the result of parsing a ConfigSpec check."""

    def __init__(self, sType, aArgs, dKwargs, sDefault):
        self.sType, self.aArgs, self.dKwargs, self.sDefault = \
            sType, aArgs, dKwargs, sDefault
        self.oEntry = self.create_widget()

    def format_value(self, oValue):
        """Return a string representation of a value."""
        return str(oValue)

    def create_widget(self):
        """Return a widget for editing this config spec."""
        raise NotImplementedError

    def set_value(self, oValue):
        """Set an editing widget value."""
        raise NotImplementedError

    def get_value(self):
        """Get the value from an editing widget."""
        raise NotImplementedError


class UneditableSpec(BaseParsedSpec):
    """Class for a spec entry that can't be edited"""

    # pylint: disable-msg=W0201
    # we define _oOrigValue outside __init__, but it's OK because create_widget
    # is called in init
    def create_widget(self):
        """Create a suitable widget (gtk.Label)"""
        self._oOrigValue = None
        return gtk.Label()

    def set_value(self, oValue):
        """Set the label value"""
        self._oOrigValue = oValue
        self.oEntry.set_text(str(oValue))

    def get_value(self):
        """Get the label value"""
        return self._oOrigValue


class StringParsedSpec(BaseParsedSpec):
    """Class for a option taking an arbitrary string"""

    def create_widget(self):
        """Create suitable widget (gtk.Entry)"""
        return gtk.Entry()

    def set_value(self, oValue):
        """Set the entry widget value"""
        if oValue is None:
            oValue = ""
        self.oEntry.set_text(str(oValue))

    def get_value(self):
        """Get the entry widget value"""
        return self.oEntry.get_text()


class BooleanParsedSpec(BaseParsedSpec):
    """Class for a Boolean option in the spec"""

    def create_widget(self):
        """Create a a suitable widget (gtk.CheckBox)"""
        return gtk.CheckButton()

    def set_value(self, oValue):
        """Set the checkbox correctly for the given value"""
        if oValue:
            self.oEntry.set_active(True)
        else:
            self.oEntry.set_active(False)

    def get_value(self):
        """Get the checkbox value"""
        return self.oEntry.get_active()


class OptionParsedSpec(BaseParsedSpec):
    """Widget for dealing with a list of exclusive options"""

    def create_widget(self):
        """Create a suitable widget (gtk.ComboBox)"""
        oCombo = gtk.combo_box_new_text()
        for sValue in self.aArgs:
            oCombo.append_text(sValue)
        return oCombo

    def set_value(self, oValue):
        """Set the values on the combo box"""
        try:
            iIndex = self.aArgs.index(oValue)
        except ValueError:
            iIndex = -1
        self.oEntry.set_active(iIndex)

    def get_value(self):
        """Get the combo box value"""
        iIndex = self.oEntry.get_active()
        if iIndex < 0 or iIndex >= len(self.aArgs):
            return None
        return self.aArgs[iIndex]


class OptionListParsedSpec(BaseParsedSpec):
    """Class for dealing with a list of non-exclusive options"""

    def create_widget(self):
        """Create suitable widget (list of gtk.CheckBoxes)"""
        oContainer = gtk.VBox()
        for sValue in self.aArgs:
            oContainer.pack_start(gtk.CheckButton(sValue))
        return oContainer

    def format_value(self, oValue):
        """Return a string representation of the value."""
        if not oValue:
            return "None"
        else:
            return ", ".join(oValue)

    def set_value(self, oValue):
        """Set the selected list correctly"""
        oValSet = set(oValue)
        for oButton in self.oEntry.get_children():
            oButton.set_active(oButton.get_label() in oValSet)

    def get_value(self):
        """Get the current list of selected items"""
        return [oButton.get_label() for oButton in self.oEntry.get_children()
            if oButton.get_active()]


class IntegerParsedSpec(BaseParsedSpec):
    """Class for dealing with integer valued options"""

    def create_widget(self):
        """Create suitable widget (gtk.Adjustment)"""
        iMin = self.dKwargs.get("min", 0)
        iMax = self.dKwargs.get("max", 100)

        oAdj = gtk.Adjustment(iMin, iMin, iMax, 1.0, 5.0, 0.0)
        oSpin = gtk.SpinButton(oAdj, 0, 0)
        oSpin.set_numeric(True)
        return oSpin

    def set_value(self, oValue):
        """Set the adjustment value, ignoring errornous input"""
        try:
            oValue = int(oValue)
        except ValueError:
            pass
        else:
            self.oEntry.set_value(oValue)

    def get_value(self):
        """Get the value form the gtk.Adjustment"""
        return self.oEntry.get_value_as_int()


SPEC_TYPE_MAP = {
    "string": StringParsedSpec,
    "boolean": BooleanParsedSpec,
    "option": OptionParsedSpec,
    "option_list": OptionListParsedSpec,
    "integer": IntegerParsedSpec,
}


def parse_spec(sConfigSpec, oValidator):
    """Parse a configobj spec into a parsed spec."""
    # TODO: is it okay to use an _ method from the validator?
    # pylint: disable-msg=W0212
    # We know we're accessing a protected member, and pylint flags our
    # note above, so don't flag this twice
    sType, aArgs, dKwargs, sDefault = \
        oValidator._parse_with_caching(sConfigSpec)
    cParsedSpec = SPEC_TYPE_MAP.get(sType, UneditableSpec)
    return cParsedSpec(sType, aArgs, dKwargs, sDefault)



# CardLookup.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2007 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Lookup AbstractCards for a list of card names.
   """

from sqlobject import SQLObjectNotFound
# pylint: disable-msg=W0402
# we need string.punctuation
import string
# pylint: enable-msg=W0402
from sutekh.core.SutekhObjects import IPhysicalCard, IExpansion, IAbstractCard
from sutekh.core.Filters import CardNameFilter, FilterAndBox, \
        make_illegal_filter


# pylint: disable-msg=R0922
# We inherit from these classes elsewhere
class LookupFailed(Exception):
    """Raised when an AbstractCard lookup fails completed.
       """
    pass


class AbstractCardLookup(object):
    """Base class for objects which translate card names into abstract card
       objects.
       """

    def lookup(self, aNames, sInfo):
        """Return a list of AbstractCards, one for each item in aNames.

           Names for which AbstractCards could not be found will be marked with
           a None in the returned list. The method may raise LookupFailed if the
           entire list should be considered invalid (e.g. if the names are
           presented to a user who then cancels the operation).
           """
        raise NotImplementedError


class PhysicalCardLookup(object):
    """Base class for objects which translate card and expansion names
       into physical card objects
       """

    def physical_lookup(self, dCardExpansions, dNameCards, dNameExps, sInfo):
        """Returns a list of physical cards. Since physical cards can't
           be repeated, this is a list of statisfable requests.

           dCardExpansions[Name][Expansion] is the number of cards requested,
           dNameCards is a dictionary of card name to abstract card mappings
           and dNameExps is a dictionary of expansion name to expansion object
           mappings.

           Note that len(list returned) =< sum(all requests in dCardExpansions)

           The physical card list will be smaller if no matching card can be
           found in the physical card list or if dAbstactCards has elements
           that have been excluded.

           LookupFailed will be raised if the entire list should be considered
           invalid, as for AbstractCardLookup
           """
        raise NotImplementedError


class ExpansionLookup(object):
    """Base class for objects which translate expansion names into expansion
       objects
       """

    def expansion_lookup(self, aExpansionNames, sInfo, dCardExpansions):
        """Return a list of Expansions, one for each item in aExpansionNames.

           Names for which Expansions could not be found will be marked with
           a None in the returned list. The method may raise LookupFailed if the
           entire list should be considered invalid (e.g. if the names are
           presented to a user who then cancels the operation).
           """
        raise NotImplementedError


class SimpleLookup(AbstractCardLookup, PhysicalCardLookup, ExpansionLookup):
    """A really straightforward lookup of AbstractCards and PhysicalCards.

       The default when we don't have a more cunning plan.
       """

    def lookup(self, aNames, _sInfo):
        """A lookup method that excludes unknown cards."""
        aCards = []
        for sName in aNames:
            if sName:
                # pylint: disable-msg=E1101
                # SQLObject confuses pylint
                try:
                    oAbs = IAbstractCard(sName)
                    aCards.append(oAbs)
                except SQLObjectNotFound:
                    aCards.append(None)
            else:
                aCards.append(None)
        return aCards

    def physical_lookup(self, dCardExpansions, dNameCards, dNameExps, _sInfo):
        """Lookup cards in the physical card set, excluding unknown cards."""
        aCards = []
        for sName in dCardExpansions:
            oAbs = dNameCards[sName]
            if oAbs is not None:
                for sExpansionName in dCardExpansions[sName]:
                    # pylint: disable-msg=W0704
                    # Do nothing exception correct here
                    try:
                        iCnt = dCardExpansions[sName][sExpansionName]
                        oExpansion = dNameExps[sExpansionName]
                        aCards.extend(
                                [IPhysicalCard((oAbs, oExpansion))] * iCnt)
                    except SQLObjectNotFound:
                        # This card is missing from the PhysicalCard list, so
                        # skipped
                        pass
        return aCards

    def expansion_lookup(self, aExpansionNames, _sInfo, _dCardExpansions):
        """Lookup for expansion names, excluding unknown expansions."""
        aExps = []
        for sExp in aExpansionNames:
            if sExp:
                try:
                    oExp = IExpansion(sExp)
                    aExps.append(oExp)
                except SQLObjectNotFound:
                    aExps.append(None)
            else:
                aExps.append(None)
        return aExps


def best_guess_filter(sName):
    """Create a filter for selecting close matches to a card name."""
    # Set the filter on the Card List to one the does a
    # Best guess search
    sFilterString = ' ' + sName.lower() + ' '
    # Kill the's in the string
    sFilterString = sFilterString.replace(' the ', ' ')
    # Kill commas, as possible issues
    sFilterString = sFilterString.replace(',', ' ')
    # Free style punctuation
    for sPunc in string.punctuation:
        sFilterString = sFilterString.replace(sPunc, '_')
    # Stolen semi-concept from soundex - replace vowels with wildcards
    # Should these be %'s ??
    # (Should at least handle the Rotscheck variation as it stands)
    sFilterString = sFilterString.replace('a', '_')
    sFilterString = sFilterString.replace('e', '_')
    sFilterString = sFilterString.replace('i', '_')
    sFilterString = sFilterString.replace('o', '_')
    sFilterString = sFilterString.replace('u', '_')
    # Normalise spaces and Wildcard spaces
    sFilterString = ' '.join(sFilterString.split())
    sFilterString = sFilterString.replace(' ', '%')
    # Add % on outside
    sFilterString = '%' + sFilterString + '%'
    oLegalFilter = make_illegal_filter()
    return FilterAndBox([CardNameFilter(sFilterString), oLegalFilter])


DEFAULT_LOOKUP = SimpleLookup()



# test_ARDBTextParser.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2007 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""Test case for ARDB text parser"""

from sutekh.io.ARDBTextParser import ARDBTextParser
from sutekh.tests.TestCore import SutekhTest
import unittest

ARDB_TEXT_EXAMPLE_1 = """
Deck Name : Test Deck
Author : Anon Y Mous
Description :
    Simple test deck.

http://www.example.url/in/description

Crypt [3 vampires] Capacity min: 2 max: 10 average: 7.33
------------------------------------------------------------
2x Test Vamp 1			  10 ...
1x Test Vamp 2			  2  ...
...

Library [19 cards]
------------------------------------------------------------
Action [6]
2x Test Card 1
4x Test Card 2

Action Modifier [12]
12x Test Card 3

Reaction [1]
1x Test Card 4
...
"""

ARDB_TEXT_EXAMPLE_2 = """
Deck Name : Test Deck 2
Author : Anon Y Mous
Description : Simple test deck.
Crypt: (3 vampires, Min: 2, Max: 10 Ave: 7.33)
------------------------------------------------------------
2 Test Vamp 1			  aus dom for   10 ...
1 Test Vamp 2			  DOM for obf   2  ...
...

Library [19 cards]
------------------------------------------------------------
Action [6]
2 Test Card 1
4 Test Card 2

Action Modifier [12]
12 Test Card 3

Reaction [1]
1 Test Card 4
...
"""

ARDB_TEXT_EXAMPLE_3 = """
Deck Name : Test Deck 2
Created By: Anon Y Mous

Crypt: (3 vampires, Min: 2, Max: 10 Ave: 7.33)
------------------------------------------------------------
2 x Test Vamp 1			  aus dom for   10 ...
1 x Test Vamp 2			  DOM for obf   2  ...
...

Library [19 cards]
------------------------------------------------------------
Action [6]
2 x Test Card 1
4 x Test Card 2

Action Modifier [12]
12 Test Card 3

Reaction [1]
1 Test Card 4
...
"""


class ARDBTextParserTests(SutekhTest):
    """class for the ARDB/FELDB text input parser"""

    def test_basic(self):
        """Run the input test."""
        oParser = ARDBTextParser()
        oHolder = self._make_holder_from_string(oParser, ARDB_TEXT_EXAMPLE_1)

        self.assertEqual(oHolder.name, "Test Deck")
        self.assertEqual(oHolder.author, "Anon Y Mous")
        self.failUnless(oHolder.comment.startswith(
            "    Simple test deck."))
        self.failUnless(oHolder.comment.endswith("in/description\n\n"))

        aCards = oHolder.get_cards()

        self.assertEqual(len(aCards), 6)
        self.failUnless(("Test Vamp 1", 2) in aCards)
        self.failUnless(("Test Vamp 2", 1) in aCards)
        self.failUnless(("Test Card 1", 2) in aCards)
        self.failUnless(("Test Card 2", 4) in aCards)
        self.failUnless(("Test Card 3", 12) in aCards)
        self.failUnless(("Test Card 4", 1) in aCards)

        oHolder = self._make_holder_from_string(oParser, ARDB_TEXT_EXAMPLE_2)

        self.assertEqual(oHolder.name, "Test Deck 2")
        self.assertEqual(oHolder.author, "Anon Y Mous")
        self.failUnless(oHolder.comment.startswith(
            "Simple test deck."))

        aCards = oHolder.get_cards()

        self.assertEqual(len(aCards), 6)
        self.failUnless(("Test Vamp 1", 2) in aCards)
        self.failUnless(("Test Vamp 2", 1) in aCards)
        self.failUnless(("Test Card 1", 2) in aCards)
        self.failUnless(("Test Card 2", 4) in aCards)
        self.failUnless(("Test Card 3", 12) in aCards)
        self.failUnless(("Test Card 4", 1) in aCards)

        oHolder = self._make_holder_from_string(oParser, ARDB_TEXT_EXAMPLE_3)

        self.assertEqual(oHolder.name, "Test Deck 2")
        self.assertEqual(oHolder.author, "Anon Y Mous")

        aCards = oHolder.get_cards()

        self.assertEqual(len(aCards), 6)
        self.failUnless(("Test Vamp 1", 2) in aCards)
        self.failUnless(("Test Vamp 2", 1) in aCards)
        self.failUnless(("Test Card 1", 2) in aCards)
        self.failUnless(("Test Card 2", 4) in aCards)
        self.failUnless(("Test Card 3", 12) in aCards)
        self.failUnless(("Test Card 4", 1) in aCards)


if __name__ == "__main__":
    unittest.main()



# DataPack.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2011 Simon Cross <hodgestar+sutekh@gmail.com>,
# Copyright 2009,2010,2011 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details
# Class and functions to manage zip data pack handling for Sutekh
# Split off from sutekh/gui/plugins/StarterDeckInfo.py and refactored,
#    June 2011

"""Provide tools for locating and extracting data pack ZIP files."""

import urllib2
import re


DOC_URL = 'http://sourceforge.net/apps/trac/sutekh/wiki/' \
          'UserDocumentation?format=txt'

ZIP_URL_BASE = 'http://sourceforge.net/apps/trac/sutekh/raw-attachment'


def find_data_pack(sTag, sDocUrl=DOC_URL, sZipUrlBase=ZIP_URL_BASE):
    """Read a documentation page and find a data pack URL.

    Returns None if no file could be found for the given tag.

    The section looked for looks something like:

    || '''Description''' || '''Tag''' || '''Date Updated''' || '''File'''
    || Some text || starters || date || [attachment:Foo.zip:wiki:FilePage Foo.zip] ||
    || Other text || rulebooks || date || [attachment:Bar.zip:wiki:FilePage Bar.zip] ||
    || Other text || twd || date || [attachment:Bas.zip:wiki:FilePage Bas.zip] ||

    dates are expected to be formated YYYY-MM-DD (strftime('%Y-%m-%d'))
    to avoid ambiguity.
    """
    oFile = urllib2.urlopen(sDocUrl)
    oHeaderRe = re.compile(r'^\|\|.*Tag')
    oAttachmentRe = re.compile(r'\[attachment:(?P<path>[^ ]*) ')
    iTagField = None
    iAttachField = None
    sZipUrl = None

    def fields(sLine):
        """Helper function to split table lines into the needed structure"""
        sLine = sLine.strip()
        return [sField.strip(" '") for sField in sLine.split('||')]

    for sLine in oFile.readlines():
        if iTagField is None:
            oMatch = oHeaderRe.match(sLine)
            if oMatch:
                aFields = fields(sLine)
                if 'Tag' not in aFields or 'File' not in aFields:
                    continue
                iTagField = aFields.index('Tag')
                iAttachField = aFields.index('File')
        else:
            aFields = fields(sLine)
            if len(aFields) > iTagField and sTag == aFields[iTagField] \
                   and len(aFields) > iAttachField:
                oMatch = oAttachmentRe.search(aFields[iAttachField])
                if oMatch:
                    sPath = oMatch.group('path')
                    # We need to split off the zip file name and the path bits
                    sZipName, sPath = sPath.split(':', 1)
                    sPath = sPath.replace(':', '/')
                    sZipUrl = '/'.join((sZipUrlBase, sPath, sZipName))

    return sZipUrl



# WriteArdbXML.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Copyright 2006 Neil Muller <drnlmuller+sutekh@gmail.com>
# Based off the Anarach Revolt Deck Builder xml support,
# ARDB (c) Francios Gombalt, Christoph Boget, Ville Virta and Vincent Ripoll
# GPL - see COPYING for details

"""Given a list of Abstract Cards in a set, write a XML file compatable with
   the Anarch Revolt Deck Builder."""

from sutekh.core.SutekhObjects import canonical_to_csv
from sutekh.core.ArdbInfo import ArdbInfo
from sutekh.io.IOBase import BaseXMLWriter
import time
# pylint: disable-msg=E0611, F0401
# xml.etree is a python2.5 thing
try:
    from xml.etree.ElementTree import Element, SubElement
except ImportError:
    from elementtree.ElementTree import Element, SubElement
# pylint: enable-msg=E0611, F0401


class WriteArdbXML(ArdbInfo, BaseXMLWriter):
    """Reformat cardset to elementTree and export it to a ARDB
       compatible XML file."""

    def _add_date_version(self, oRoot):
        """Add the standard data to the root element"""
        sDateWritten = time.strftime('%Y-%m-%d', time.localtime())
        oRoot.attrib['generator'] = "Sutekh [ %s ]" % self.sVersionString
        oRoot.attrib['formatVersion'] = self.sFormatVersion
        oRoot.attrib['databaseVersion'] = self.sDatabaseVersion
        oDateElem = SubElement(oRoot, 'date')
        oDateElem.text = sDateWritten

    # pylint: disable-msg=R0201
    # Methods so they're available to the subclasses
    def _ardb_crypt_card(self, oCardElem, oAbsCard, sSet):
        """Fill in name, set and advanced elements for a crypt card"""
        oAdvElem = SubElement(oCardElem, 'adv')
        oNameElem = SubElement(oCardElem, 'name')
        sName = canonical_to_csv(oAbsCard.name)
        if oAbsCard.level is not None:
            oAdvElem.text = 'Advanced'
            # This is a bit hackish
            oNameElem.text = sName.replace(' (Advanced)', '')
        else:
            oNameElem.text = sName
        oSetElem = SubElement(oCardElem, 'set')
        oSetElem.text = sSet

    def _ardb_lib_card(self, oCardElem, oAbsCard, sSet):
        """Fill in name and set for a library card"""
        oNameElem = SubElement(oCardElem, 'name')
        oNameElem.text = canonical_to_csv(oAbsCard.name)
        oSetElem = SubElement(oCardElem, 'set')
        oSetElem.text = sSet

    # pylint: enable-msg=R0201

    def _gen_tree(self, oHolder):
        """Creates the actual XML document into memory."""
        dCards = self._get_cards(oHolder.cards)
        dVamps, dCryptStats = self._extract_crypt(dCards)
        dLib, iLibSize = self._extract_library(dCards)
        oRoot = Element('deck')

        self._add_date_version(oRoot)

        oNameElem = SubElement(oRoot, 'name')
        oNameElem.text = oHolder.name
        oAuthElem = SubElement(oRoot, 'author')
        oAuthElem.text = oHolder.author
        oDescElem = SubElement(oRoot, 'description')
        oDescElem.text = oHolder.comment

        oCryptElem = SubElement(oRoot, 'crypt', size=str(dCryptStats['size']),
                min=str(dCryptStats['min']), max=str(dCryptStats['max']),
                avg='%.2f' % dCryptStats['avg'])
        self.format_vamps(oCryptElem, dVamps)
        oLibElem = SubElement(oRoot, 'library', size=str(iLibSize))
        self.format_library(oLibElem, dLib)
        return oRoot

    def format_vamps(self, oCryptElem, dVamps):
        """Convert the Vampire dictionary into ElementTree representation."""
        for (oCard, sSet), iNum in sorted(dVamps.iteritems(),
                key=lambda x: (x[0][0].name, x[0][1], x[1])):
            # This won't match the ARDB ID's, unless by chance.
            # It looks like that should not be an issue as ARDB will
            # use the name if the IDs don't match
            oCardElem = SubElement(oCryptElem, 'vampire',
                    databaseID=str(oCard.id), count=str(iNum))
            self._ardb_crypt_card(oCardElem, oCard, sSet)
            oDiscElem = SubElement(oCardElem, 'disciplines')
            oDiscElem.text = self._gen_disciplines(oCard)
            oClanElem = SubElement(oCardElem, 'clan')
            oCapElem = SubElement(oCardElem, 'capacity')
            if len(oCard.creed) > 0:
                # ARDB seems to treat all Imbued as being of the same clan
                # Should we do an Imbued:Creed thing?
                oClanElem.text = 'Imbued'
                oCapElem.text = str(oCard.life)
            else:
                oClanElem.text = [x.name for x in oCard.clan][0]
                oCapElem.text = str(oCard.capacity)
            oGrpElem = SubElement(oCardElem, 'group')
            oGrpElem.text = str(oCard.group)
            # ARDB doesn't handle sect specifically
            # No idea how ARDB represents independant titles -
            # this seems set when the ARDB database is created, and is
            # not in the ARDB codebase
            if len(oCard.title) > 0:
                oTitleElem = SubElement(oCardElem, 'title')
                oTitleElem.text = [oC.name for oC in oCard.title][0]
            oTextElem = SubElement(oCardElem, 'text')
            oTextElem.text = oCard.text

    def format_library(self, oLibElem, dLib):
        """Format the dictionary of library cards for the element tree."""
        for (oCard, sTypeString, sSet), iNum in sorted(dLib.iteritems(),
                key=lambda x: (x[0][0].name, x[0][1], x[1])):
            oCardElem = SubElement(oLibElem, 'card', databaseID=str(oCard.id),
                    count=str(iNum))
            self._ardb_lib_card(oCardElem, oCard, sSet)
            if oCard.costtype is not None:
                oCostElem = SubElement(oCardElem, 'cost')
                oCostElem.text = "%d %s" % (oCard.cost, oCard.costtype)
            if len(oCard.clan) > 0:
                # ARDB also strores things like "requires a prince"
                # we don't, so too bad
                oReqElem = SubElement(oCardElem, 'requirement')
                oReqElem.text = ".".join([x.name for x in oCard.clan])
            oTypeElem = SubElement(oCardElem, 'type')
            oTypeElem.text = sTypeString
            # Not sure if this does quite the right thing here
            if len(oCard.discipline) > 0:
                oDiscElem = SubElement(oCardElem, 'disciplines')
                oDiscElem.text = self._gen_disciplines(oCard)
            oTextElem = SubElement(oCardElem, 'text')
            oTextElem.text = oCard.text



# DBSignals.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# Define the reload singal we need
# Copyright 2007 Simon Cross <hodgestar@gmail.com>,
# Copyright 2007 Neil Muller <drnlmuller+sutekh@gmail.com>
# GPL - see COPYING for details

"""Wrappers around SQLObject signals needed to keep card sets and the card
collection in sync."""

from sqlobject.events import Signal, listen, RowUpdateSignal, \
        RowDestroySignal, RowCreatedSignal
from sqlobject.include.pydispatch import dispatcher
from sutekh.core.SutekhObjects import PhysicalCardSet


class ChangedSignal(Signal):
    """Syncronisation signal for card sets.

       Needs to be sent after changes are commited to the database, so card
       sets can reload properly.
       Used so card sets always reflect correct available counts.
       """


# Senders
def send_changed_signal(oCardSet, oPhysCard, iChange, cClass=PhysicalCardSet):
    """Sent when card counts change, as card sets may need to update."""
    cClass.sqlmeta.send(ChangedSignal, oCardSet, oPhysCard, iChange)


# Listeners
def listen_changed(fListener, cClass):
    """Listens for the changed_signal."""
    listen(fListener, cClass, ChangedSignal)


def listen_row_destroy(fListener, cClass):
    """listen for the row destroyed signal sent when a card set is deleted."""
    listen(fListener, cClass, RowDestroySignal)


def listen_row_update(fListener, cClass):
    """listen for the row updated signal sent when a card set is modified."""
    listen(fListener, cClass, RowUpdateSignal)


def listen_row_created(fListener, cClass):
    """listen for the row updated signal sent when a card set is modified."""
    listen(fListener, cClass, RowCreatedSignal)


def disconnect_changed(fListener, cClass):
    """Disconnects from the changed_signal."""
    dispatcher.disconnect(fListener, signal=ChangedSignal, sender=cClass)


def disconnect_row_destroy(fListener, cClass):
    """Disconnect from the row destroyed signal."""
    dispatcher.disconnect(fListener, signal=RowDestroySignal, sender=cClass)


def disconnect_row_created(fListener, cClass):
    """Disconnect from the row destroyed signal."""
    dispatcher.disconnect(fListener, signal=RowCreatedSignal, sender=cClass)


def disconnect_row_update(fListener, cClass):
    """Disconnect the row updated signal."""
    dispatcher.disconnect(fListener, signal=RowUpdateSignal, sender=cClass)



# RulingParser.py
# -*- coding: utf8 -*-
# vim:fileencoding=utf8 ai ts=4 sts=4 et sw=4
# WhiteWolf Rulings Parser
# Copyright 2005, 2006 Simon Cross <hodgestar@gmail.com>
# GPL - see COPYING for details

"""HTML Parser for extracting card rulings from the WW online rulings list."""

import re
from logging import Logger
from sutekh.io.SutekhBaseHTMLParser import SutekhBaseHTMLParser, StateError, \
        LogState, LogStateWithInfo
from sutekh.core.SutekhObjects import IAbstractCard, SutekhObjectMaker
from sqlobject import SQLObjectNotFound


# Ruling Saver
class RuleDict(dict):
    """Dictionary object which holds the extracted rulings information."""

    _aSectionKeys = ['title', 'card']
    _oMasterOut = re.compile(r'\s*-\s*Master\s*\:?\s*Out\-of\-Turn$')
    _oCommaThe = re.compile(r'\s*\,\s*The$')
    _dOddTitles = {
        'Absilmilard\'s Army': 'Absimiliard\'s Army',
        'Carlotta': 'Carlotta Giovanni',
        'Donal O\'Connor': u'D\xf3nal O\'Connor',
        'Herald of Topeth': 'Herald of Topheth',
        'Illusions of Kindred': 'Illusions of the Kindred',
        'Khobar Towers': 'Khobar Towers, Al-Khubar',
        'Mehemet of the Ahl-i-Batin': 'Mehemet of the Ahl-i-Batin (Mage)',
        'Lazar Dobrescu': u'L\xe1z\xe1r Dobrescu',
        'Merill Molitor': 'Merrill Molitor',
        'Rotschreck': u'R\xf6tschreck',
        'Seattle Committe': 'Seattle Committee',
        'Shackles of Enkindu': 'Shackles of Enkidu',
        'Shadow Court Satyr': 'Shadow Court Satyr (Changeling)',
        'Smiling Jack the Anarch': 'Smiling Jack, The Anarch',
        'Tereza Rotas': 'Tereza Rostas',
        'Ur-Shulgi': 'Ur-Shulgi, The Shepherd',
    }

    def __init__(self, oLogger):
        self.oLogger = oLogger
        super(RuleDict, self).__init__()
        self._oMaker = SutekhObjectMaker()

    def _find_card(self, sTitle):
        """Find the abstract card this rules applies to."""
        sTitle = self._oMasterOut.sub('', sTitle)
        sTitle = self._oCommaThe.sub('', sTitle)

        # pylint: disable-msg=W0704
        # Skipping SQLObject exceptions is the right thing to do here
        try:
            return IAbstractCard(sTitle)
        except SQLObjectNotFound:
            pass

        try:
            return IAbstractCard(self._dOddTitles[sTitle])
        except KeyError:
            pass
        except SQLObjectNotFound:
            pass

        try:
            return IAbstractCard(('The ' + sTitle))
        except SQLObjectNotFound:
            pass

        return None

    def clear_rule(self):
        """Remove current contents of the rule."""
        dKeep = {}
        for sKey in self._aSectionKeys:
            if self.has_key(sKey):
                dKeep[sKey] = self[sKey]
        self.clear()
        for (sKey, sValue) in dKeep.items():
            self[sKey] = sValue

    def save(self):
        """Add the rulings to the database."""
        if not (self.has_key('title') and self.has_key('code') \
                and self.has_key('text')):
            return

        if not self.has_key('card'):
            self['card'] = self._find_card(self['title'])

        if self['card'] is None:
            return

        self.oLogger.info('Card: %s', self['card'].name)

        oRuling = self._oMaker.make_ruling(self['text'], self['code'])

        if self.has_key('url'):
            oRuling.url = self['url']

        self['card'].addRuling(oRuling)


# State Classes
class NoSection(LogState):
    """Not in any ruling section."""

    def transition(self, sTag, _dAttr):
        """Transition to InSection if needed."""
        if sTag == 'p':
            return InSection(RuleDict(self.oLogger), self.oLogger)
        else:
            return self


class InSection(LogStateWithInfo):
    """In a ruling section."""

    def transition(self, sTag, _dAttr):
        """Transition to SectionTitle or NoSection as needed."""
        if sTag == 'b':
            return SectionTitle(self._dInfo, self.oLogger)
        elif sTag == 'p':
            # skip to next section
            return InSection(RuleDict(self.oLogger), self.oLogger)
        else:
            return NoSection(self.oLogger)


class SectionTitle(LogStateWithInfo):
    """In the title of the section."""

    def transition(self, sTag, _dAttr):
        """Transition to SectionWithTitle if needed."""
        if sTag == 'b':
            raise StateError()
        elif sTag == '/b':
            self._dInfo['title'] = self._sData.strip().strip(':')
            return SectionWithTitle(self._dInfo, self.oLogger)
        else:
            return self


class SectionWithTitle(LogStateWithInfo):
    """In a section with a known title."""

    def transition(self, sTag, _dAttr):
        """Transition to SectionRule, InSection or NoSection if needed."""
        if sTag == 'ul':
            return SectionRule(self._dInfo, self.oLogger)
        elif sTag == 'p':
            # skip to next section
            return InSection(RuleDict(self.oLogger), self.oLogger)
        elif sTag == '/ul':
            return NoSection(self.oLogger)
        else:
            return self


class SectionRule(LogStateWithInfo):
    """In a ruling in the section."""

    def transition(self, sTag, _dAttr):
        """Transition to the appropriate InRule State."""
        if sTag == 'li':
            return InRuleText(self._dInfo, self.oLogger)
        elif sTag == '/ul':
            return NoSection(self.oLogger)
        else:
            return self


class InRuleText(LogStateWithInfo):
    """In the text of a ruling."""

    oCodePattern = re.compile('\[[^]]*\]')

    def transition(self, sTag, dAttr):
        """Transition to SectionRule if needed."""
        if sTag == 'a':
            if not self._dInfo.has_key('text'):
                self._dInfo['text'] = self._sData.strip()
            self._dInfo['url'] = dAttr['href']
            return InRuleUrl(self._dInfo, self.oLogger)
        elif sTag == '/li':
            if not self._dInfo.has_key('code'):
                # Rule without an url, so extract the last section in []'s
                # from the text
                oMatch = self.oCodePattern.search(self._sData)
                if oMatch:
                    # Add code, and remove it from text
                    self._dInfo['code'] = oMatch.group()
                    self._dInfo['text'] = \
                            self.oCodePattern.sub('', self._sData).strip()
            self._dInfo.save()
            self._dInfo.clear_rule()
            return SectionRule(self._dInfo, self.oLogger)
        return self


class InRuleUrl(LogStateWithInfo):
    """In the url associated with this ruling."""

    def transition(self, sTag, _dAttr):
        """Transition to SectionRule if needed."""
        if sTag == 'a':
            raise StateError()
        elif sTag == '/a':
            # Some rulings have multiple codes, but we only take the last one
            # since that matches the url we keep
            self._dInfo['code'] = self._sData.strip()
            return InRuleText(self._dInfo, self.oLogger)
        else:
            return self


# Parser
class RulingParser(SutekhBaseHTMLParser):
    """Actual Parser for the WW rulings HTML files."""

    def __init__(self, oLogHandler):
        # super().__init__ calls reset, so we need this first
        self.oLogger = Logger('WW Rulings parser')
        if oLogHandler is not None:
            self.oLogger.addHandler(oLogHandler)
        super(RulingParser, self).__init__()
        # No need to touch self._oState, reset will do that for us

    def reset(self):
        """Reset the parser"""
        super(RulingParser, self).reset()
        self._oState = NoSection(self.oLogger)
